"""
Base kupfer objects definitions.

This file is a part of the program kupfer, which is
released under GNU General Public License v3 (or any later version),
see the main program file, and COPYING for details.
"""

from __future__ import annotations

import time
import typing as ty

from kupfer import icons
from kupfer.support import itertools, kupferstring, pretty

if ty.TYPE_CHECKING:
    from gettext import gettext as _

    from gi.repository import GdkPixbuf

    from kupfer.core import commandexec

__all__ = [
    "Action",
    "ActionGenerator",
    "AnySource",
    "KupferObject",
    "Leaf",
    "Source",
    "TextSource",
    "_NonpersistentToken",
]


class KupferObject:
    """Base class for kupfer data model

    This class provides a way to get at an object's:

    * icon with get_thumbnail, get_pixbuf and get_icon
    * name with unicode() or str()
    * description with get_description

    @rank_adjust should be used _very_ sparingly:
        Default actions should have +5 or +1
        Destructive (dangerous) actions should have -5 or -10

    @fallback_icon_name is a class attribute for the last fallback
    icon; it must always be accessible.
    """

    rank_adjust: int = 0
    fallback_icon_name: str = "kupfer-object"
    # _is_builtin: bool = False

    def __init__(self, name: str | None = None) -> None:
        """Init kupfer object with."""
        self.name: str = name or self.__class__.__name__
        folded_name = kupferstring.tofolded(self.name)
        self.kupfer_add_alias(folded_name)

    def kupfer_add_alias(self, alias: str) -> None:
        """Add alias to object. This create name_aliases set if not exist
        for this leaf."""
        if alias != str(self):
            if not hasattr(self, "name_aliases"):
                self.name_aliases = set()

            self.name_aliases.add(alias)

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        if cached := getattr(self, "_cached_repr", None):
            return ty.cast("str", cached)

        if key := self.repr_key():
            rep = f"<{self.__module__}.{self.__class__.__name__} {key}>"
        else:
            rep = f"<{self.__module__}.{self.__class__.__name__}>"

        self._cached_repr = rep
        return rep

    def repr_key(self) -> ty.Any:
        """Return an object whose str() will be used in the __repr__,
        self is returned by default.  This value is used to recognize objects,
        for example learning commonly used objects."""
        return self

    def get_description(self) -> str | None:
        """Return a description of the specific item."""
        return None

    def get_thumbnail(
        self, width: int, height: int
    ) -> GdkPixbuf.Pixbuf | None:
        """Return pixbuf of size @width x @height if available.
        Most objects will not implement this."""
        return None

    def get_pixbuf(self, icon_size: int) -> GdkPixbuf.Pixbuf | None:
        """Returns an icon in pixbuf format with dimension @icon_size.

        Subclasses should implement: get_gicon and get_icon_name,
        if they make sense.  The methods are tried in that order."""

        if (gicon := self.get_gicon()) and (
            pbuf := icons.get_icon_for_gicon(gicon, icon_size)
        ):
            return pbuf

        if (icon_name := self.get_icon_name()) and (
            icon := icons.get_icon_for_name(icon_name, icon_size)
        ):
            return icon

        return icons.get_icon_for_name(self.fallback_icon_name, icon_size)

    def get_icon(self) -> icons.GIcon | None:
        """Returns an icon in GIcon format.

        Subclasses should implement: get_gicon and get_icon_name,
        if they make sense. The methods are tried in that order."""

        if (gicon := self.get_gicon()) and icons.is_good_gicon(gicon):
            return gicon

        if (
            icon_name := self.get_icon_name()
        ) and icons.get_good_name_for_icon_names((icon_name,)):
            return icons.get_gicon_for_names(icon_name)

        return icons.get_gicon_for_names(self.fallback_icon_name)

    def get_gicon(self) -> icons.GIcon | None:
        """Return GIcon, if there is one"""
        return None

    def get_icon_name(self) -> str:
        """Return icon name. All items should have at least a generic icon name
        to return."""
        return self.fallback_icon_name


T = ty.TypeVar("T")


class _NonpersistentToken(ty.Generic[T]):
    """Hold data that goes None when pickled."""

    __slots__ = ("object",)

    def __init__(self, object_: T) -> None:
        self.object: T = object_

    def __bool__(self) -> bool:
        return bool(self.object)

    def __reduce__(self):
        return (sum, ((), None))


class Leaf(KupferObject):
    """Base class for objects

    Leaf.object is the represented object (data)
    All Leaves should be hashable (__hash__ and __eq__)
    """

    def __init__(self, obj: ty.Any, name: str) -> None:
        """Represented object @obj and its @name"""
        super().__init__(name)
        self.object = obj
        self._content_source: _NonpersistentToken[Source] | None = None

    def __hash__(self) -> int:
        return hash(str(self))

    def __eq__(self, other: object) -> bool:
        return type(self) is type(other) and self.object == other.object  # type: ignore

    def add_content(self, content: Source | None) -> None:
        """Register content source @content with Leaf"""
        if content:
            self._content_source = _NonpersistentToken(content)

    def has_content(self) -> bool:
        return bool(self._content_source)

    def content_source(self, alternate: bool = False) -> Source | None:
        """Content of leaf. it MAY alter behavior with @alternate,
        as easter egg/extra mode"""
        if self._content_source:
            return self._content_source.object

        return None

    def get_actions(self) -> ty.Iterable[Action]:
        """Default (builtin) actions for this Leaf"""
        return ()


class Action(KupferObject):
    '''Base class for all actions.

    Implicit interface:
      valid_object will be called once for each (secondary) object
      to see if it applies. If it is not defined, all objects are
      assumed ok (within the other type/source constraints)

    def valid_object(self, obj, for_item):
        """Whether @obj is good for secondary obj,
        where @for_item is passed in as a hint for
        which it should be applied to
        """
        return True

    @action_accelerator: str or None
        Default single lowercase letter key to use for selecting the action
        quickly
    '''

    fallback_icon_name: str = "kupfer-execute"
    action_accelerator: str | None = None

    def __hash__(self) -> int:
        return hash(repr(self))

    def __eq__(self, other: object) -> bool:
        return (
            type(self) is type(other)
            and repr(self) == repr(other)
            and str(self) == str(other)
        )

    def repr_key(self) -> ty.Any:
        """by default, actions of one type are all the same"""
        return None

    def activate(
        self,
        leaf: Leaf,
        iobj: Leaf | None = None,
        ctx: commandexec.ExecutionToken | None = None,
    ) -> Leaf | None:
        """Use this action with @obj and @iobj

        leaf: the direct object (Leaf)
        iobj: the indirect object (Leaf), if ``self.requires_object``
              returns ``False``
        ctx:  optional ExecutionToken

        if ``self.wants_context`` returns ``True``, then the action
        also receives an execution context object as ``ctx``.

        Also, ``activate_multiple(self, objects, iobjects=None, ctx=None)``
        is called if it is defined and the action gets either
        multiple objects or iobjects.
        """
        raise NotImplementedError

    def wants_context(self) -> bool:
        """Return ``True`` if ``activate`` should receive the
        ActionExecutionContext as the keyword argument context

        Defaults to ``False`` in accordance with the old protocol
        """
        return False

    def is_factory(self) -> bool:
        """Return whether action may return a result collection as a Source"""
        return False

    def has_result(self) -> bool:
        """Return whether action may return a result item as a Leaf"""
        return False

    def is_async(self) -> bool:
        """If this action runs asynchronously, return True.

        Then activate(..) must return an object from the kupfer.task module,
        which will be queued to run by Kupfer's task scheduler.
        """
        return False

    def item_types(self) -> ty.Iterable[ty.Type[Leaf]]:
        """Yield types this action may apply to. This is used only
        when this action is specified in __kupfer_actions__ to "decorate"
        """
        return ()

    def valid_for_item(self, leaf: Leaf) -> bool:
        """Whether action can be used with exactly @item"""
        return True

    def requires_object(self) -> bool:
        """If this action requires a secondary object to complete is action,
        return True."""
        return False

    def object_source(self, for_item: Leaf | None = None) -> Source | None:
        """Source to use for object or None, to use the catalog (flat and
        filtered for @object_types)."""
        return None

    def object_source_and_catalog(self, for_item: Leaf) -> bool:
        return False

    def object_types(self) -> ty.Iterable[ty.Type[Leaf]]:
        """Yield types this action may use as indirect objects, if the action
        requrires it."""
        return ()


class Source(KupferObject, pretty.OutputMixin):
    """Source: Data provider for a kupfer browser.

    All Sources should be hashable and treated as equal if
    their @repr are equal!

    *source_user_reloadable*
        if True source get "Reload" action without debug mode.

    *source_prefer_sublevel*
        if True, the source by default exports its contents in a subcatalog, not
        to the toplevel.  NOTE: *Almost never* use this: let the user decide,
        default to toplevel.

    *source_use_cache*
        if True, the source can be pickled to disk to save its
        cached items until the next time the launcher is started.

    *source_scan_interval*
        set typical rescan interval (not guaranteed) in seconds.
        Set 0 to use default. values lower than min_rescan_interval in
        PeriodicRescanner are ignored. Warning: last scan time is pickled, so
        after restart - source may be immediately rescanned.

    *rank_adjust*
        change all leaves rank by given value. This only apply to search with
        key when given leaf already pass minimum rank level.
    """

    fallback_icon_name = "kupfer-object-multiple"
    source_user_reloadable = False
    source_prefer_sublevel = False
    source_use_cache = True
    source_scan_interval: int = 0

    def __init__(self, name):
        KupferObject.__init__(self, name)
        self.cached_items: ty.Iterable[Leaf] | None = None
        self._version: int = 1
        # last source rescan timestamp
        self.last_scan: int = 0

    @property
    def version(self) -> int:
        """Version is for pickling (save and restore from cache),
        subclasses should increase self._version when changing."""
        return self._version

    def __eq__(self, other):
        return (
            type(self) is type(other)
            and repr(self) == repr(other)
            and self.version == other.version
        )

    def __hash__(self) -> int:
        return hash(repr(self))

    def toplevel_source(self) -> Source:
        return self

    def initialize(self) -> None:
        """Called when a Source enters Kupfer's system for real.

        This method is called at least once for any "real" Source. A Source
        must be able to return an icon name for get_icon_name as well as a
        description for get_description, even if this method was never called.
        """

    def finalize(self) -> None:
        """Called before a source is deactivated."""

    def repr_key(self) -> ty.Any:
        return None

    def get_items(self) -> ty.Iterable[Leaf]:
        """Internal method to compute and return the needed items.

        Subclasses should use this method to return a sequence or
        iterator to the leaves it contains
        """
        return []

    def get_items_forced(self) -> ty.Iterable[Leaf]:
        """Force compute and return items for source.
        Default - call get_items method."""
        return self.get_items()

    def is_dynamic(self) -> bool:
        """Whether to recompute contents each time it is accessed."""
        return False

    def mark_for_update(self, postpone: bool = False) -> None:
        """Mark source as changed.

        When postpone is True, cached items are not cleaned so be available to
        use and refreshed on next PeriodicRescanner run.

        If there is no cached_items source load items on next use.
        """
        self.last_scan = 0
        if not postpone:
            self.cached_items = None

    def should_sort_lexically(self) -> bool:
        """Sources should return items by most relevant order (most relevant
        first). If this is True, Source will sort items from get_item()
        in locale lexical order."""
        return False

    def get_leaves(self, force_update: bool = False) -> ty.Iterable[Leaf]:
        """Return a list of leaves.

        Subclasses should implement ``get_items()``, so that ``Source`` can
        handle sorting and caching.

        If *force_update*, ignore cache, load *all* leaves from source and store
        it into cache. Print number of items loaded.

        If source ``get_items()`` return list or tuple - also all leaves are
        loaded and stored in cache.

        Otherwise (when ``get_items()`` is generator) leaves are loaded on
        request and cached in ``SavedIterable``.

        If ``should_sort_lexically()`` is True - all leaves are loaded, sorted
        and cached.
        """

        if self.is_dynamic():
            items = (
                self.get_items_forced() if force_update else self.get_items()
            )

            if self.should_sort_lexically():
                items = kupferstring.locale_sort(items)

            return items

        if self.cached_items is None or force_update:
            items = (
                self.get_items_forced() if force_update else self.get_items()
            )
            if self.should_sort_lexically():
                # sorting make list
                items = kupferstring.locale_sort(items)

            if force_update:
                # make list when needed
                self.cached_items = itertools.as_list(items)
                self.output_debug(f"Loaded {len(self.cached_items)} items")
            elif isinstance(items, (list, tuple)):
                self.cached_items = items
                self.output_debug(f"Loaded {len(items)} items (l)")
            else:
                # use savediterable only for iterators
                self.cached_items = itertools.SavedIterable(items)
                self.output_debug("Loaded items")

            self.last_scan = int(time.time())

        return self.cached_items or ()

    def has_parent(self) -> bool:
        """Return True when source has other, parent Source."""
        return False

    def get_parent(self) -> Source | None:
        """Return parent Source for given source if exists or None."""
        return None

    def get_leaf_repr(self) -> Leaf | None:
        """Return, if applicable, another object to take the source's place as
        Leaf"""
        return None

    def get_valid_leaf_repr(self) -> tuple[bool, Leaf | None]:
        """Return, if applicable, another object to take the source's place as
        Leaf.  Return tuple (leaf representation is valid, leaf).
        Valid representation may be None, so first element of tuple must be
        checked."""
        if (
            (leaf_repr := self.get_leaf_repr())
            and (hasattr(leaf_repr, "is_valid"))
            and not leaf_repr.is_valid()
        ):
            return False, None

        return True, leaf_repr

    def provides(self) -> ty.Iterable[ty.Type[Leaf]]:
        """A seq of the types of items it provides; empty is taken as anything
        -- however most sources should set this to exactly the type they yield.
        """
        return ()

    def get_search_text(self) -> str:
        """Message displayed on gui when no item is selected."""
        return _("Type to search")

    def get_empty_text(self) -> str:
        """Text displayed when no leaves found in source."""
        return _("%s is empty") % str(self)


class TextSource(KupferObject):
    """TextSource base class implementation, this is a pseudo Source."""

    def __init__(
        self,
        name: str | None = None,
        placeholder: str | None = None,
    ) -> None:
        """
        *name*:         Localized name
        *placeholder*:  Localized placeholder when it has no input
        """
        if not name:
            name = _("Text")

        KupferObject.__init__(self, name)
        self.placeholder = placeholder

    def __eq__(self, other: object) -> bool:
        return type(self) is type(other) and repr(self).__eq__(repr(other))

    def __hash__(self) -> int:
        return hash(repr(self))

    def initialize(self) -> None:
        pass

    def repr_key(self) -> ty.Any:
        return None

    def get_rank(self) -> int:
        """All items are given this rank."""
        return 20

    def get_items(self, text: str) -> ty.Iterable[Leaf]:
        return ()

    def get_text_items(self, text: str) -> ty.Iterable[Leaf]:
        """Get leaves for string `text`."""
        return self.get_items(text)

    def has_parent(self) -> bool:
        return False

    def provides(self) -> ty.Iterable[ty.Type[Leaf]]:
        """A seq of the types of items it provides"""
        yield Leaf

    def get_icon_name(self) -> str:
        return "edit-select-all"

    def get_search_text(self) -> str:
        return self.placeholder or _("Text")

    def get_empty_text(self) -> str:
        return self.placeholder or _("Text")


# pylint: disable=too-few-public-methods
class ActionGenerator:
    """A "source" for actions.

    NOTE: The ActionGenerator should not perform any expensive
    computation, and not access any slow media (files, network) when
    returning actions.  Such expensive checks must be performed in
    each Action's valid_for_item method.
    """

    def get_actions_for_leaf(self, leaf: Leaf) -> ty.Iterable[Action]:
        """Return actions appropriate for given leaf."""
        return ()


AnySource = ty.Union[Source, TextSource]



from __future__ import annotations

import shutil
import typing as ty

from gi.repository import GObject

from kupfer.core import plugins, settings
from kupfer.support import fileutils, pretty

if ty.TYPE_CHECKING:
    from gettext import gettext as _
    from gettext import ngettext

__all__ = [
    "PluginSettings",
    "UserNamePassword",
    "check_any_command_available",
    "check_command_available",
    "check_dbus_connection",
    "check_keyring_support",
    "register_alternative",
]


def _is_core_setting(key: str) -> bool:
    return key.startswith("kupfer_")


SettingChangeCB = ty.Callable[[ty.Any, str, ty.Any], None]


class PluginSettings(GObject.GObject, pretty.OutputMixin):  # type:ignore
    """Allows plugins to have preferences by assigning an instance
    of this class to the plugin's __kupfer_settings__ attribute.

    Setting values are accessed by the getitem operator [] with
    the setting's 'key' attribute

    Signals:

        plugin-setting-changed: key, value

    """

    __gtype_name__ = "PluginSettings"

    def __init__(self, *setdescs: dict[str, ty.Any]) -> None:
        """Create a settings collection by passing in dictionaries
        as arguments, where each dictionary must have the following keys:
            key
            type
            value (default value)
            label (localized label)

        the @key may be any string except strings starting with
        'kupfer_', which are reserved
        """
        GObject.GObject.__init__(self)
        self.setting_descriptions: dict[str, dict[str, ty.Any]] = {}
        self.setting_key_order: list[str] = []
        self.signal_connection: int = -1
        req_keys = {"key", "value", "type", "label"}
        for desc in setdescs:
            if not req_keys.issubset(list(desc.keys())):
                missing = req_keys.difference(list(desc.keys()))
                raise KeyError(f"Plugin setting missing keys: {missing}")

            if _is_core_setting(desc["key"]):
                raise KeyError(f"Reserved plugin setting key: {desc['key']!r}")

            self.setting_descriptions[desc["key"]] = desc.copy()
            self.setting_key_order.append(desc["key"])

    def __iter__(self) -> ty.Iterator[str]:
        return iter(self.setting_key_order)

    def initialize(self, plugin_name: str) -> None:
        """Init by reading from global settings and setting up callbacks"""
        setctl = settings.get_settings_controller()
        for key in self:
            value_type = self.setting_descriptions[key]["type"]
            value = setctl.get_plugin_config(plugin_name, key, value_type)
            if value is not None:
                self[key] = value

            elif _is_core_setting(key):
                default = self.setting_descriptions[key]["value"]
                setctl.set_plugin_config(plugin_name, key, default, value_type)

        setctl.connect("value-changed", self._on_value_changed, plugin_name)
        # register for unload notification
        if not plugin_name.startswith("core."):
            plugins.register_plugin_unimport_hook(
                plugin_name, self._disconnect_all, plugin_name
            )

    def __getitem__(self, key: str) -> ty.Any:
        return self.setting_descriptions[key]["value"]

    def __setitem__(self, key: str, value: ty.Any) -> None:
        value_type = self.setting_descriptions[key]["type"]
        self.setting_descriptions[key]["value"] = value_type(value)
        if not _is_core_setting(key):
            self.emit(f"plugin-setting-changed::{key}", key, value)

    def _on_value_changed(
        self,
        setctl: settings.SettingsController,
        section: str,
        key: str,
        value: ty.Any,
        plugin_name: str,
    ) -> None:
        """Preferences changed, update object"""
        if key in self and plugin_name in section:
            self[key] = value

    def get_value_type(self, key: str) -> ty.Type[ty.Any]:
        """Return type of setting @key"""
        return ty.cast(
            "ty.Type[ty.Any]", self.setting_descriptions[key]["type"]
        )

    def get_label(self, key: str) -> str:
        """Return label for setting @key"""
        return ty.cast("str", self.setting_descriptions[key]["label"])

    def get_alternatives(self, key: str) -> ty.Iterable[ty.Any] | None:
        """Return alternatives for setting @key (if any)"""
        return self.setting_descriptions[key].get("alternatives")

    def get_tooltip(self, key: str) -> str | None:
        """Return tooltip string for setting @key (if any)"""
        return self.setting_descriptions[key].get("tooltip")

    def get_parameter(
        self, key: str, name: str, default: ty.Any = None
    ) -> ty.Any:
        return self.setting_descriptions[key].get(name, default)

    def connect_settings_changed_cb(
        self, callback: SettingChangeCB, *args: ty.Any
    ) -> None:
        self.signal_connection = self.connect(
            "plugin-setting-changed", callback, *args
        )

    def _disconnect_all(self, plugin_name: str) -> None:
        if self.signal_connection != -1:
            self.disconnect(self.signal_connection)


# Arguments: Key, Value
# Detailed by the key
GObject.signal_new(
    "plugin-setting-changed",
    PluginSettings,
    GObject.SignalFlags.RUN_LAST | GObject.SignalFlags.DETAILED,
    GObject.TYPE_BOOLEAN,
    (GObject.TYPE_STRING, GObject.TYPE_PYOBJECT),
)

# Plugin convenience functions for dependencies


# pylint: disable=too-few-public-methods
class _DBusChecker:
    has_connection: bool | None = None

    @classmethod
    def check(cls) -> bool:
        if cls.has_connection is None:
            try:
                import dbus  # pylint: disable=import-outside-toplevel

                dbus.Bus()
                cls.has_connection = True
            except (ImportError, dbus.DBusException):
                cls.has_connection = False

        assert cls.has_connection is not None
        return cls.has_connection


def check_dbus_connection() -> None:
    """Check if a connection to the D-Bus daemon is available,
    else raise ImportError with an explanatory error message.

    For plugins that can not be used without contact with D-Bus;
    if this check is used, the plugin may use D-Bus and assume it
    is available in the Plugin's code.
    """
    if not _DBusChecker.check():
        raise ImportError(_("No D-Bus connection to desktop session"))


# pylint: disable=too-few-public-methods
class UserNamePassword:
    pass


def check_keyring_support() -> None:
    """raise ImportError with because it is not supported."""
    raise ImportError("Keyring is not supported")


def check_keybinding_support() -> None:
    """Check if we can make global keybindings."""
    from kupfer.ui import (
        keybindings,  # pylint: disable=import-outside-toplevel
    )

    if not keybindings.is_available():
        raise ImportError(_("Dependency '%s' is not available") % "Keybinder")


def _plugin_configuration_error(plugin: str, err: ty.Any) -> None:
    pretty.print_error(__name__, err)


def _is_valid_terminal(term_dict: dict[str, ty.Any]) -> bool:
    if len(term_dict["argv"]) < 1:
        return False

    exe = term_dict["argv"][0]
    return bool(fileutils.lookup_exec_path(exe))


_AVAILABLE_ALTERNATIVES: ty.Final[dict[str, dict[str, ty.Any]]] = {
    "terminal": {
        "filter": _is_valid_terminal,
        "required_keys": {
            "name": str,
            "argv": list,
            "exearg": str,
            "desktopid": str,
            "startup_notify": bool,
        },
    },
    "editor": {
        "filter": None,
        "required_keys": {
            "name": str,
            "argv": list,
            "terminal": bool,
        },
    },
    "icon_renderer": {
        "filter": None,
        "required_keys": {
            "name": str,
            "renderer": object,
        },
    },
}

_ALTERNATIVES: dict[str, dict[str, ty.Any]] = {
    "terminal": {},
    "editor": {},
    "icon_renderer": {},
}


def register_alternative(
    caller: str, category_key: str, id_: str, **kwargs: ty.Any
) -> bool:
    """Register a new alternative for the category @category_key.

    @caller: Must be the caller's plugin id (Plugin __name__ variable)

    @id_ is a string identifier for the object to register
    @kwargs are the keyed arguments for the alternative constructor

    Returns True with success
    """
    if category_key not in _AVAILABLE_ALTERNATIVES:
        _plugin_configuration_error(
            caller, f"Category '{category_key}' does not exist"
        )
        return False

    full_id = f"{caller}.{id_}"
    kw_set = set(kwargs)
    alt = _AVAILABLE_ALTERNATIVES[category_key]
    req_set = set(alt["required_keys"])

    if full_id in _ALTERNATIVES[category_key]:
        pretty.print_debug(
            __name__,
            f"Alternative {full_id} already defined in {category_key}",
        )
        return False

    if not req_set.issubset(kw_set):
        _plugin_configuration_error(
            caller, f"Configuration error for alternative '{category_key}':"
        )
        _plugin_configuration_error(
            caller, f"Missing keys: {req_set - kw_set}"
        )
        return False

    _ALTERNATIVES[category_key][full_id] = kwargs
    pretty.print_debug(
        __name__, f"Registered alternative {category_key}: {full_id}"
    )
    setctl = settings.get_settings_controller()
    setctl.update_alternatives(
        category_key, _ALTERNATIVES[category_key], alt["filter"]
    )

    # register the alternative to be unloaded
    plugin_id = ".".join(caller.split(".")[2:])
    if plugin_id and not plugin_id.startswith("core."):
        plugins.register_plugin_unimport_hook(
            plugin_id, _unregister_alternative, caller, category_key, full_id
        )

    return True


def _unregister_alternative(
    caller: str, category_key: str, full_id: str
) -> None:
    """Remove the alternative for category @category_key
    (this is done automatically at plugin unload)
    """
    if category_key not in _AVAILABLE_ALTERNATIVES:
        _plugin_configuration_error(
            caller, f"Category '{category_key}' does not exist"
        )
        return

    alt = _AVAILABLE_ALTERNATIVES[category_key]
    try:
        del _ALTERNATIVES[category_key][full_id]
    except KeyError:
        _plugin_configuration_error(
            caller, f"Alternative '{full_id}' does not exist"
        )
        return

    pretty.print_debug(
        __name__, f"Unregistered alternative {category_key}: {full_id}"
    )
    setctl = settings.get_settings_controller()
    setctl.update_alternatives(
        category_key, _ALTERNATIVES[category_key], alt["filter"]
    )


def check_command_available(*cmd: str) -> None:
    """Check if the commands is available in system, throw ImportError when not"""
    missing = [f'"{c}"' for c in cmd if not shutil.which(c)]
    if not missing:
        return

    raise ImportError(
        ngettext(
            "Command %(msg)s is not available.",
            "Commands %(msg)s are not available.",
            len(missing),
        )
        % {"msg": ", ".join(missing)}
    )


def check_any_command_available(*cmd: str) -> None:
    """Check if any of command is available in system, throw ImportError when not"""
    for c in cmd:
        if shutil.which(c):
            return

    raise ImportError(
        _("None of commands %(msg)s is available.")
        % {"msg": ", ".join(f'"{c}"' for c in cmd)}
    )



import contextlib
import traceback
import typing as ty

from kupfer.core import plugins
from kupfer.core.plugins import (
    PluginAttr,
    initialize_plugin,
    load_plugin_objects,
)
from kupfer.support import pretty

if ty.TYPE_CHECKING:
    from kupfer.obj.base import Action, ActionGenerator, Source, TextSource

__all__ = ("exception_guard", "load_plugin", "remove_plugin")


# pylint: disable=too-few-public-methods
class PluginDescription:
    def __init__(self):
        self.text_sources: list[TextSource] = []
        self.action_decorators: list[Action] = []
        self.content_decorators: list[ty.Type[Source]] = []
        self.action_generators: list[ActionGenerator] = []
        self.sources: list[Source] = []


def load_plugin(plugin_id: str) -> PluginDescription:
    """Load plugin by `plugin_id`.

    @S_sources are to be included directly in the catalog,
    @s_sources as just as subitems
    """
    sources: list[Source] = []
    text_sources: list[TextSource] = []
    action_decorators: list[Action] = []
    content_decorators: list[ty.Type[Source]] = []
    action_generators: list[ActionGenerator] = []

    initialize_plugin(plugin_id)
    if not plugins.is_plugin_loaded(plugin_id):
        return PluginDescription()

    text_sources.extend(
        load_plugin_objects(plugin_id, PluginAttr.TEXT_SOURCES)
    )
    action_decorators.extend(
        load_plugin_objects(plugin_id, PluginAttr.ACTION_DECORATORS)
    )
    action_generators.extend(
        load_plugin_objects(plugin_id, PluginAttr.ACTION_GENERATORS)
    )

    # Register all Sources as (potential) content decorators
    content_decorators.extend(
        load_plugin_objects(plugin_id, PluginAttr.SOURCES, instantiate=False)
    )
    content_decorators.extend(
        load_plugin_objects(
            plugin_id, PluginAttr.CONTENT_DECORATORS, instantiate=False
        )
    )
    sources.extend(load_plugin_objects(plugin_id, PluginAttr.SOURCES))

    desc = PluginDescription()

    desc.text_sources = text_sources
    desc.action_decorators = action_decorators
    desc.content_decorators = content_decorators
    desc.sources = sources
    desc.action_generators = action_generators
    return desc


@contextlib.contextmanager
def exception_guard(name, *args, callback=None, **kwargs):
    "Guard for exceptions, print traceback and call @callback if any is raised"
    try:
        yield
    except Exception:
        pretty.print_error(__name__, f"Loading {name} raised an exception:")
        traceback.print_exc()
        pretty.print_error(__name__, "This error is probably a bug in", name)
        pretty.print_error(__name__, "Please file a bug report")
        if callback is not None:
            callback(*args, **kwargs)


def remove_plugin(plugin_id: str) -> None:
    plugins.unimport_plugin(plugin_id)



"""
Stub implementation of HTTPS connections.
"""


# pylint: disable=too-few-public-methods
class VerifiedHTTPSConnection:
    "implementation stub"

    def __init__(self, host, *args, **kwargs):
        pass

    @classmethod
    def is_ssl_supported(cls):
        return False


def is_supported():
    return (
        VerifiedHTTPSConnection  # type: ignore
        and VerifiedHTTPSConnection.is_ssl_supported()
    )



from __future__ import annotations

__kupfer_name__ = _("Applications")
__kupfer_sources__ = ("AppSource",)
__kupfer_actions__ = (
    "OpenWith",
    "OpenWithByMime",
    "SetDefaultApplication",
    "ResetAssociations",
    "LaunchHere",
    "AppAdditionalAction",
)
__description__ = _("All applications and preferences")
__version__ = "2023.1"
__author__ = ""

import typing as ty
from pathlib import Path

from gi.repository import Gio

from kupfer import config, plugin_support
from kupfer.obj import Action, AppLeaf, FileLeaf, Leaf, Source, UrlLeaf
from kupfer.obj.helplib import FilesystemWatchMixin
from kupfer.support import weaklib

if ty.TYPE_CHECKING:
    from gettext import gettext as _


_ALTERNATIVES = (
    "",
    "Cinnamon",
    "EDE",
    "GNOME",
    "KDE",
    "LXDE",
    "LXQt",
    "MATE",
    "Pantheon",
    "ROX",
    "Razor",
    "TDE",
    "Unity",
    "XFCE",
)

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "desktop_type",
        "label": _("Applications for Desktop Environment"),
        "type": str,
        "value": "",
        "alternatives": _ALTERNATIVES,
    },
    {
        "key": "desktop_filter",
        "label": _("Use Desktop Filter"),
        "type": bool,
        "value": True,
    },
    {
        "key": "load_extra_aliases",
        "label": _("Load extra aliases"),
        "type": bool,
        "value": False,
        "tooltip": _(
            "Load additional aliases like keywords or "
            "other names. This may slowdown searching. "
            "Change this setting may require reload."
        ),
    },
)

# Gio.AppInfo / Desktop Item nodisplay vs hidden:
# NoDisplay: Don't show this in program menus
# Hidden: Disable/never use at all

WHITELIST_IDS: ty.Final = (
    # we think that these are useful to show
    "eog.desktop",
    "evince.desktop",
    "gnome-about.desktop",
    "gstreamer-properties.desktop",
    "notification-properties.desktop",
    "shotwell-viewer.desktop",
)
BLACKLIST_IDS: ty.Final = ("nautilus-home.desktop",)


def initialize_plugin(name: str) -> None:
    def on_change_settings(*args: ty.Any) -> None:
        AppLeaf.LOAD_EXTRA_ALIASES = __kupfer_settings__["load_extra_aliases"]

    __kupfer_settings__.connect_settings_changed_cb(on_change_settings)


def _should_show(
    app_info: Gio.AppInfo, desktop_type: str, use_filter: bool
) -> bool:
    if app_info.get_nodisplay():
        return False

    if not use_filter:
        return True

    if desktop_type == "":
        return app_info.should_show()  # type: ignore

    return app_info.get_show_in(desktop_type)  # type:ignore


class AppSource(Source, FilesystemWatchMixin):
    """Applications source

    This Source contains all user-visible applications (as given by
    the desktop files).
    """

    source_scan_interval: int = 3600

    def __init__(self, name=None):
        super().__init__(name or _("Applications"))
        self.monitor_token = None

    def initialize(self):
        application_dirs = config.get_data_dirs("", "applications")
        self.monitor_token = self.monitor_directories(*application_dirs)
        weaklib.gobject_connect_weakly(
            __kupfer_settings__,
            "plugin-setting-changed",
            self._on_setting_change,
        )

    def _on_setting_change(self, *_args):
        self.mark_for_update()

    def get_items(self):
        use_filter = __kupfer_settings__["desktop_filter"]
        desktop_type = __kupfer_settings__["desktop_type"]

        # Add this to the default
        for item in Gio.app_info_get_all():
            id_ = item.get_id()
            if id_ in WHITELIST_IDS or (
                _should_show(item, desktop_type, use_filter)
                and id_ not in BLACKLIST_IDS
            ):
                # load extra app action defined in desktop file
                app_actions = [
                    (action, item.get_action_name(action))
                    for action in item.list_actions()
                ]

                yield AppLeaf(item, app_actions=app_actions)

    def should_sort_lexically(self):
        return True

    def get_description(self):
        return _("All applications and preferences")

    def get_icon_name(self):
        return "applications-office"

    def provides(self):
        yield AppLeaf


class OpenWith(Action):
    action_accelerator = "w"

    def __init__(self):
        super().__init__(_("Open With Any Application..."))

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx
        assert iobj
        self.activate_multiple((leaf,), (iobj,), ctx)

    def activate_multiple(
        self,
        objects: ty.Iterable[Leaf],
        iobjects: ty.Iterable[Leaf],
        ctx: ty.Any,
    ) -> None:
        # for each application, launch all the files
        paths = [L.object for L in objects]
        for iobj_app in iobjects:
            assert isinstance(iobj_app, AppLeaf)
            iobj_app.launch(paths=paths, ctx=ctx)

    def item_types(self):
        yield FileLeaf

    def requires_object(self):
        return True

    def object_types(self):
        yield AppLeaf

    def object_source(self, for_item=None):
        return AppsAll()

    def object_source_and_catalog(self, for_item):
        return True

    def valid_object(self, iobj, for_item):
        return iobj.object.supports_files() or iobj.object.supports_uris()

    def get_description(self):
        return _("Open with any application")


class OpenWithByMime(Action):
    rank_adjust = 5
    action_accelerator = "w"

    def __init__(self):
        super().__init__(_("Open With..."))

    def wants_context(self):
        return True

    def activate(
        self, leaf: Leaf, iobj: Leaf | None = None, ctx: ty.Any = None
    ) -> None:
        assert ctx
        assert iobj
        self.activate_multiple((leaf,), (iobj,), ctx)

    def activate_multiple(
        self,
        objects: ty.Iterable[Leaf],
        iobjects: ty.Iterable[Leaf],
        ctx: ty.Any,
    ) -> None:
        # for each application, launch all the files
        files: list[Gio.File] = []
        files.extend(
            Gio.File.new_for_path(p.object)
            for p in objects
            if not isinstance(p, UrlLeaf)
        )
        files.extend(
            Gio.File.new_for_uri(p.object)
            for p in objects
            if isinstance(p, UrlLeaf)
        )

        for iobj_app in iobjects:
            assert isinstance(iobj_app, AppLeaf)
            iobj_app.launch(files=files, ctx=ctx)

    def item_types(self):
        yield FileLeaf
        yield UrlLeaf

    def object_types(self):
        yield AppLeaf

    def requires_object(self):
        return True

    def object_source(self, for_item=None):
        if isinstance(for_item, FileLeaf) and (
            mime := for_item.get_content_type()
        ):
            return AppsForMime(mime)

        return None

    def valid_object(self, iobj, for_item):
        return iobj.object.supports_files() or iobj.object.supports_uris()

    def get_description(self):
        return _("Open with application supporting this file type")


class SetDefaultApplication(Action):
    def __init__(self):
        super().__init__(_("Set Default Application..."))

    def activate(self, leaf, iobj=None, ctx=None):
        assert iobj
        desktop_item = iobj.object
        desktop_item.set_as_default_for_type(leaf.get_content_type())

    def item_types(self):
        yield FileLeaf

    def requires_object(self):
        return True

    def object_types(self):
        yield AppLeaf

    def object_source(self, for_item=None):
        return AppsAll()

    def object_source_and_catalog(self, for_item):
        return True

    def valid_object(self, iobj, for_item):
        return iobj.object.supports_files() or iobj.object.supports_uris()

    def get_description(self):
        return _("Set default application to open this file type")


class AppsAll(Source):
    def __init__(self):
        super().__init__(_("Applications"))

    def initialize(self):
        weaklib.gobject_connect_weakly(
            __kupfer_settings__,
            "plugin-setting-changed",
            self._on_setting_change,
        )

    def _on_setting_change(self, *_args):
        self.mark_for_update()

    def get_items(self):
        use_filter = __kupfer_settings__["desktop_filter"]
        desktop_type = __kupfer_settings__["desktop_type"]

        # Get all apps; this includes those only configured for
        # opening files with.
        for item in Gio.AppInfo.get_all():
            if _should_show(item, desktop_type, use_filter):
                continue

            if not item.supports_uris() and not item.supports_files():
                continue

            yield AppLeaf(item)

    def should_sort_lexically(self):
        return False

    def get_description(self):
        return None

    def get_icon_name(self):
        return "applications-office"

    def provides(self):
        yield AppLeaf


class AppsForMime(Source):
    source_use_cache = False

    def __init__(self, mimetype: str) -> None:
        super().__init__(_("Applications supporting %s") % mimetype)
        self._mimetype = mimetype

    def get_items(self) -> ty.Iterator[AppLeaf]:
        for item in Gio.app_info_get_all_for_type(self._mimetype):
            if not item.supports_uris() and not item.supports_files():
                continue

            yield AppLeaf(item)

    def should_sort_lexically(self):
        return False

    def provides(self):
        yield AppLeaf


class ResetAssociations(Action):
    rank_adjust = -10

    def __init__(self):
        super().__init__(_("Reset Associations"))

    def activate(self, leaf, iobj=None, ctx=None):
        content_type = leaf.get_content_type()
        Gio.AppInfo.reset_type_associations(content_type)

    def item_types(self):
        yield FileLeaf

    def get_description(self):
        return _("Reset program associations for files of this type.")


class LaunchHere(Action):
    def __init__(self, name=_("Start Application Here...")):
        super().__init__(name)

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx
        assert iobj
        if leaf.is_dir():
            work_dir = leaf.object
        else:
            work_dir = str(Path(leaf.object).parent)

        iobj.launch(activate=False, ctx=ctx, work_dir=work_dir)

    def get_description(self) -> str:
        return _("Launch application in this location")

    def get_icon_name(self) -> str:
        return "kupfer-launch"

    def item_types(self):
        yield FileLeaf

    # def valid_for_item(self, leaf):
    #     # leaf is FileLeaf
    #     return leaf.is_dir()

    def requires_object(self):
        return True

    def object_types(self):
        yield AppLeaf

    def object_source(self, for_item=None):
        return AppsAll()

    def object_source_and_catalog(self, for_item):
        return True


class _AppAction(Leaf):
    def __init__(self, action: str, name: str) -> None:
        super().__init__(action, name)

    def get_icon_name(self) -> str:
        return "kupfer-launch"


class _AppActionsSource(Source):
    source_use_cache = False

    def __init__(self, appleaf: AppLeaf) -> None:
        super().__init__("application actions")
        self.appleaf = appleaf

    def get_items(self):
        if self.appleaf.app_actions:
            for action, name in self.appleaf.app_actions:
                yield _AppAction(action, name)

    def provides(self):
        yield _AppAction


class AppAdditionalAction(Action):
    def __init__(self, name=_("Application actions...")):
        super().__init__(name)

    def activate(self, leaf, iobj=None, ctx=None):
        assert iobj
        assert isinstance(leaf, AppLeaf)
        assert leaf.app_actions
        appinfo = leaf.object
        appinfo.launch_action(iobj.object)

    def item_types(self):
        yield AppLeaf

    def requires_object(self):
        return True

    def object_types(self):
        yield _AppAction

    def object_source(self, for_item=None):
        return _AppActionsSource(for_item)

    def valid_for_item(self, leaf: Leaf) -> bool:
        # application must have any actions
        assert isinstance(leaf, AppLeaf)
        return bool(leaf.app_actions)

    def get_description(self):
        return _("Extra actions for application")

    def get_icon_name(self) -> str:
        return "gtk-execute"



from __future__ import annotations

import typing as ty

from gi.repository import Gio

from kupfer.obj import FileLeaf
from kupfer.support import scheduler

if ty.TYPE_CHECKING:
    from kupfer.core.commandexec import ExecutionToken

__all__ = ("AsyncFileResult", "register_async_file_result")


# pylint: disable=too-few-public-methods
class AsyncFileResult:
    """Expect a given file path to be created, and when (probably) done,
    post the file as a late result."""

    def __init__(
        self, ctx: ExecutionToken, filepath: str, timeout: int = 0
    ) -> None:
        self.ctx = ctx
        gfile = Gio.File.new_for_path(filepath)
        self.monitor = gfile.monitor_file(Gio.FileMonitorFlags.NONE)
        self.callback_id = self.monitor.connect("changed", self._on_changed)
        self._cancel_timer = None
        if timeout:
            self._cancel_timer = scheduler.Timer(True)
            self._cancel_timer.set(timeout, self._cancel)

    def _cancel(self, *_args: ty.Any) -> None:
        """Stop waiting for file."""
        self.monitor.disconnect(self.callback_id)
        self.monitor.cancel()
        self.monitor = None
        if self._cancel_timer:
            self._cancel_timer.invalidate()

    def _on_changed(
        self,
        monitor: Gio.FileMonitor,
        gfile1: Gio.File,
        gfile2: Gio.File | None,
        event: Gio.FileMonitorEvent,
    ) -> None:
        if event == Gio.FileMonitorEvent.CHANGES_DONE_HINT:
            self.ctx.register_late_result(FileLeaf(gfile1.get_path()))
            self._cancel()


def register_async_file_result(
    ctx: ExecutionToken, filepath: str, timeout: int = 600
) -> AsyncFileResult:
    """Register that `filepath` may appear soon.
    `ctx` is the action's execution context token.
    `timeout` is maximum time to wait for file.
    """
    return AsyncFileResult(ctx, filepath, timeout)



__kupfer_name__ = _("Librewolf Bookmarks")
__kupfer_sources__ = ("BookmarksSource",)
__kupfer_actions__ = ()
__description__ = _(
    "Index of Librewolf bookmarks. "
    "Bookmark is always opened in default browser."
)
__version__ = "2023.1"
__author__ = "Ulrik, William Friesen, Karol Będkowski"

import typing as ty

from kupfer import plugin_support
from kupfer.obj import Source, UrlLeaf
from kupfer.obj.apps import AppLeafContentMixin
from kupfer.obj.helplib import FilesystemWatchMixin
from kupfer.plugin._firefox_support import (
    get_bookmarks,
    get_librewolf_home_file,
)

if ty.TYPE_CHECKING:
    from gettext import gettext as _


__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "profile",
        "label": _("Librewolf profile name or path"),
        "type": str,
        "value": "",
        "helper": "choose_directory",
    },
)


class BookmarksSource(AppLeafContentMixin, Source, FilesystemWatchMixin):
    appleaf_content_id = "librewolf"
    source_scan_interval: int = 3600

    def __init__(self):
        self.monitor_token = None
        super().__init__(_("Librewolf Bookmarks"))
        self._version = 3

    def initialize(self):
        if ff_home := get_librewolf_home_file(""):
            self.monitor_token = self.monitor_directories(str(ff_home))

    def monitor_include_file(self, gfile):
        return gfile and gfile.get_basename() == "lock"

    def mark_for_update(self, postpone=False):
        super().mark_for_update(postpone=True)

    def _get_ffx3_bookmarks(self):
        """Query the librewolf places bookmark database"""
        fpath = get_librewolf_home_file(
            "places.sqlite", __kupfer_settings__["profile"]
        )

        return get_bookmarks(fpath)

    def get_items(self):
        return self._get_ffx3_bookmarks()

    def get_description(self):
        return _("Index of Librewolf bookmarks")

    def get_gicon(self):
        if lrepr := self.get_leaf_repr():
            return lrepr.get_gicon()

        return None

    def get_icon_name(self):
        return "web-browser"

    def provides(self):
        yield UrlLeaf



"""
This module contains Helper constructs

This file is a part of the program kupfer, which is
released under GNU General Public License v3 (or any later version),
see the main program file, and COPYING for details.
"""

from __future__ import annotations

import typing as ty
from pathlib import Path

from gi.repository import Gio, GLib

from kupfer.support import pretty

if ty.TYPE_CHECKING:
    from kupfer.core import commandexec
    from kupfer.obj import base

__all__ = (
    "NonpersistentToken",
    "PicklingHelperMixin",
    "reverse_action",
)


class PicklingHelperMixin:
    """This pickling helper will define __getstate__/__setstate__
    acting simply on the class dictionary; it is up to the inheriting
    class to set up:

    pickle_prepare:
        Modify the instance dict to remove any unpickleable attributes,
        the resulting dict will be pickled
    unpickle_finish:
        Finish unpickling by restoring nonpickled attributes from the
        saved class dict, or setting up change callbacks or similar
    """

    def pickle_prepare(self) -> None:
        pass

    def unpickle_finish(self) -> None:
        pass

    def __getstate__(self) -> dict[str, ty.Any]:
        """On pickle, getstate will call self.pickle_prepare(),
        then it will return the class' current __dict__."""
        self.pickle_prepare()
        return self.__dict__

    def __setstate__(self, state: dict[str, ty.Any]) -> None:
        """On unpickle, setstate will restore the class' __dict__,
        then call self.unpickle_finish()."""
        self.__dict__.update(state)
        self.unpickle_finish()


TokenDataT = ty.TypeVar("TokenDataT")


class NonpersistentToken(PicklingHelperMixin, ty.Generic[TokenDataT]):
    """A token will keep a reference until pickling, when it is deleted"""

    data: TokenDataT | None

    def __init__(self, data: TokenDataT):
        self.data = data

    def __bool__(self) -> bool:
        return bool(self.data)

    def pickle_prepare(self) -> None:
        self.data = None


FileMonitorToken = NonpersistentToken[list[Gio.FileMonitor]]


class FilesystemWatchMixin:
    """A mixin for Sources watching directories"""

    def monitor_files(self, *files: str | Path) -> FileMonitorToken:
        """Start monitoring `files` for changes.
        Similar `monitor_directories`, but monitor also not existing files."""
        tokens = []
        for file in files:
            if isinstance(file, Path):
                file = str(file)  # noqa: PLW2901

            gfile = Gio.File.new_for_path(file)
            try:
                monitor = gfile.monitor_file(Gio.FileMonitorFlags.NONE, None)
            except GLib.GError as exc:
                pretty.print_debug(__name__, "FilesystemWatchMixin", exc)
                continue

            if monitor:
                monitor.connect("changed", self._on_file_changed)
                tokens.append(monitor)

        return NonpersistentToken(tokens)

    def monitor_directories(
        self, *directories: str | Path, force: bool = False
    ) -> FileMonitorToken:
        """Register @directories for monitoring;

        On changes, the Source will be marked for update.
        This method returns a monitor token that has to be
        stored for the monitor to be active.

        The token will be a false value if nothing could be monitored.

        Nonexisting directories are skipped, if not passing `force` True.
        """
        tokens = []
        for directory in directories:
            if isinstance(directory, Path):
                directory = str(directory)  # noqa: PLW2901

            gfile = Gio.File.new_for_path(directory)
            if not force and not gfile.query_exists():
                continue

            try:
                monitor = gfile.monitor_directory(
                    Gio.FileMonitorFlags.WATCH_MOUNTS, None
                )
            except GLib.GError as exc:
                pretty.print_debug(__name__, "FilesystemWatchMixin", exc)
                continue

            if monitor:
                monitor.connect("changed", self._on_directory_changed)
                tokens.append(monitor)

        return NonpersistentToken(tokens)

    def monitor_include_file(self, gfile: Gio.File) -> bool:
        """Return whether @gfile should trigger an update event
        by default, files beginning with "." are ignored.

        This method is used for monitoring both files and directories.
        """
        return not (gfile and gfile.get_basename().startswith("."))

    def stop_monitor_fs_changes(
        self, nptoken: FileMonitorToken | None
    ) -> None:
        """Stop monitoring for files or directories changes"""
        if nptoken and nptoken.data:
            for token in nptoken.data:
                assert isinstance(token, Gio.FileMonitor)
                token.cancel()

    def _on_directory_changed(
        self,
        _monitor: ty.Any,
        file1: Gio.File,
        _file2: ty.Any,
        evt_type: Gio.FileMonitorEvent,
    ) -> None:
        if evt_type in (
            Gio.FileMonitorEvent.CREATED,
            Gio.FileMonitorEvent.DELETED,
        ) and self.monitor_include_file(file1):
            pretty.print_debug(
                __name__, "_on_directory_changed", file1.get_path()
            )
            assert hasattr(self, "mark_for_update")
            self.mark_for_update()

    def _on_file_changed(
        self,
        _monitor: ty.Any,
        file1: Gio.File,
        _file2: ty.Any,
        evt_type: Gio.FileMonitorEvent,
    ) -> None:
        if evt_type in (
            Gio.FileMonitorEvent.CREATED,
            Gio.FileMonitorEvent.DELETED,
            Gio.FileMonitorEvent.CHANGED,
        ) and self.monitor_include_file(file1):
            pretty.print_debug(__name__, "_on_file_changed", file1.get_path())
            assert hasattr(self, "mark_for_update")
            self.mark_for_update()


def reverse_action(
    action: ty.Type[base.Action], rank: int = 0
) -> ty.Type[base.Action]:
    """Return a reversed version a three-part action.

    @action: the action class
    @rank: the rank_adjust to give the reversed action

    A three-part action requires a direct object (item) and an indirect
    object (iobj).

    In general, the item must be from the Catalog, while the iobj can be
    from one, specified special Source. If this is used, and the action
    will be reversed, the base action must be the one specifying a
    source for the iobj. The reversed action will always take both item
    and iobj from the Catalog, filtered by type.

    If valid_object(iobj, for_leaf=None) is used, it will always be
    called with only the new item as the first parameter when reversed.
    """

    class ReverseAction(action):  # type: ignore
        rank_adjust = rank

        def activate(
            self,
            leaf: base.Leaf,
            iobj: base.Leaf | None = None,
            ctx: commandexec.ExecutionToken | None = None,
        ) -> ty.Any:
            assert iobj
            return action.activate(self, iobj, leaf, ctx)

        def item_types(self) -> ty.Iterable[ty.Type[base.Leaf]]:
            return action.object_types(self)

        def valid_for_item(self, leaf: base.Leaf) -> bool:
            try:
                return action.valid_object(self, leaf)  # type: ignore
            except AttributeError:
                return True

        def object_types(self) -> ty.Iterable[ty.Type[base.Leaf]]:
            return action.item_types(self)

        def valid_object(
            self, obj: base.Leaf, for_item: base.Leaf | None = None
        ) -> bool:
            return action.valid_for_item(self, obj)

        def object_source(
            self, for_item: base.Leaf | None = None
        ) -> base.Source | None:
            return None

    ReverseAction.__name__ = "Reverse" + action.__name__
    return ReverseAction



__kupfer_name__ = _("Quick Image Viewer")
__kupfer_actions__ = ("View",)
__description__ = ""
__version__ = ""
__author__ = ""

import shutil
import typing as ty

from gi.repository import Gdk, GdkPixbuf, GLib, Gtk

from kupfer import launch
from kupfer.obj import Action, FileLeaf, OperationError

if ty.TYPE_CHECKING:
    from gettext import gettext as _


def _on_size_prepared(loader, width, height, max_w, max_h):
    if width <= max_w and height <= max_h:
        return

    w_scale = max_w * 1.0 / width
    h_scale = max_h * 1.0 / height
    scale = min(w_scale, h_scale)
    loader.set_size(int(width * scale), int(height * scale))


def load_image_max_size(
    filename: str, width: int, height: int
) -> GdkPixbuf.Pixbuf:
    pbloader = GdkPixbuf.PixbufLoader()
    pbloader.connect("size-prepared", _on_size_prepared, width, height)
    try:
        with open(filename, "rb") as fin:
            shutil.copyfileobj(fin, pbloader)

        pbloader.close()

    except (GLib.GError, OSError) as exc:
        raise OperationError(exc) from exc

    return pbloader.get_pixbuf()


class View(Action):
    def __init__(self):
        Action.__init__(self, _("View Image"))
        self.open_windows = {}

    def item_types(self):
        yield FileLeaf

    def valid_for_item(self, leaf):
        return leaf.is_content_type("image/*")

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx

        ## If the same file @obj is already open,
        ## then close its window.
        if leaf.object in self.open_windows:
            open_window = self.open_windows.pop(leaf.object)
            open_window.destroy()
            return

        image_widget = Gtk.Image()
        image_widget.set_from_pixbuf(
            load_image_max_size(
                leaf.object, Gdk.Screen.height(), Gdk.Screen.width()
            )
        )
        image_widget.show()

        window = Gtk.Window()
        window.set_title(str(leaf))
        # pylint: disable=no-member
        window.set_position(Gtk.WindowPosition.CENTER)
        # pylint: disable=no-member
        window.add(image_widget)
        ctx.environment.present_window(window)
        window.connect(
            "key-press-event", self._on_window_key_press, leaf.object
        )
        window.connect("delete-event", self._on_window_deleted, leaf.object)
        self.open_windows[leaf.object] = window

    def _on_window_deleted(self, window, event, filename):
        self.open_windows.pop(filename, None)
        return False

    def _on_window_key_press(self, window, event, filepath):
        if Gdk.keyval_name(event.keyval) == "Escape":
            self._on_window_deleted(window, event, filepath)
            window.destroy()
            return True

        if Gdk.keyval_name(event.keyval) == "Return":
            self._on_window_deleted(window, event, filepath)
            launch.show_path(filepath)
            window.destroy()
            return True

        return False

    def get_description(self):
        return None



__kupfer_name__ = _("Image Tools")
__kupfer_sources__ = ()
__kupfer_text_sources__ = ()
__kupfer_actions__ = ("Scale", "Rotate90", "Rotate270", "Autorotate")
__description__ = _(
    "Image transformation tools using 'convert' from ImageMagick"
)
__version__ = "2017.1"
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import subprocess
import typing as ty
from contextlib import suppress

# since "path" is a very generic name, you often forget..
from os import path as os_path

from kupfer import launch, plugin_support, runtimehelper
from kupfer.obj import Action, FileLeaf, OperationError, TextLeaf
from kupfer.support import fileutils, pretty

if ty.TYPE_CHECKING:
    from gettext import gettext as _


# TODO: use imagemagick -auto-orient ??

plugin_support.check_any_command_available("convert", "jpegtran", "jhead")


def _spawn_operation_err(argv):
    try:
        launch.spawn_async_raise(argv)
    except launch.SpawnError as exc:
        raise OperationError(exc.args[0].message) from exc


class Scale(Action):
    def __init__(self):
        Action.__init__(self, _("Scale..."))

    def has_result(self):
        return True

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert iobj
        size = self._make_size(iobj.object)
        fpath = leaf.object
        dirname = os_path.dirname(fpath)
        head, ext = os_path.splitext(os_path.basename(fpath))
        filename = f"{head}_{size}{ext}"
        dpath = fileutils.get_destpath_in_directory(dirname, filename)
        argv = ["convert", "-scale", str(size), fpath, dpath]
        runtimehelper.register_async_file_result(ctx, dpath)
        _spawn_operation_err(argv)
        return FileLeaf(dpath)

    def item_types(self):
        yield FileLeaf

    def valid_for_item(self, leaf):
        return leaf.is_content_type("image/*")

    def requires_object(self):
        return True

    def object_types(self):
        yield TextLeaf

    @classmethod
    def _make_size(cls, text):
        size = None
        # Allow leading =
        text = text.strip("= ")
        try:
            size = str(float(text.strip()))
        except ValueError:
            with suppress(ValueError):
                twoparts = text.split("x", 1)
                xval = float(twoparts[0].strip())
                yval = float(twoparts[1].strip())
                size = f"{xval:g}x{yval:g}"

        return size

    def valid_object(self, obj, for_item=None):
        return self._make_size(obj.object)

    def get_description(self):
        return _("Scale image to fit inside given pixel measure(s)")


class RotateBase(Action):
    def __init__(self, name, rotation):
        Action.__init__(self, name)
        self.rotation = rotation

    def has_result(self):
        return True

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx

        fpath = leaf.object
        dirname = os_path.dirname(fpath)
        head, ext = os_path.splitext(os_path.basename(fpath))
        filename = f"{head}_{self.rotation}{ext}"
        dpath = fileutils.get_destpath_in_directory(dirname, filename)
        argv = [
            "jpegtran",
            "-copy",
            "all",
            "-rotate",
            self.rotation,
            "-outfile",
            dpath,
            fpath,
        ]
        runtimehelper.register_async_file_result(ctx, dpath)
        _spawn_operation_err(argv)
        return FileLeaf(dpath)

    def item_types(self):
        yield FileLeaf

    def valid_for_item(self, leaf):
        return leaf.is_content_type("image/*")


class Rotate90(RotateBase):
    def __init__(self):
        RotateBase.__init__(self, _("Rotate Clockwise"), "90")

    def get_icon_name(self):
        return "object-rotate-right"


class Rotate270(RotateBase):
    def __init__(self):
        RotateBase.__init__(self, _("Rotate Counter-Clockwise"), "270")

    def get_icon_name(self):
        return "object-rotate-left"


class Autorotate(Action):
    def __init__(self):
        Action.__init__(self, _("Autorotate"))

    def has_result(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        fpath = leaf.object
        argv = ["jhead", "-autorot", fpath]
        launch.spawn_async(argv)

    def item_types(self):
        yield FileLeaf

    def valid_for_item(self, leaf):
        _root, ext = os_path.splitext(leaf.object)
        if ext.lower() not in (".jpeg", ".jpg"):
            return False
        # Launch jhead to see if 1) it is installed, 2) Orientation nondefault
        try:
            cmdargs = ("jhead", leaf.object)
            with subprocess.Popen(cmdargs, stdout=subprocess.PIPE) as proc:
                out, _err = proc.communicate()
                pretty.print_debug(__name__, "Running", cmdargs)
                return any(
                    li.decode("UTF-8").startswith("Orientation")
                    for li in out.splitlines()
                )

        except OSError:
            pretty.print_debug(__name__, f"Action {self} needs 'jhead'")

        return False

    def get_description(self):
        return _("Rotate JPEG (in-place) according to its EXIF metadata")



"""
Various, common types for typechecing
"""

import types
import typing as ty

ExecInfo = ty.Union[
    tuple[
        ty.Type[BaseException], BaseException, ty.Optional[types.TracebackType]
    ],
    tuple[None, None, None],
]



"""
Changes:
    2012-10-16 Karol Będkowski:
        + support Gnome3; closes lp#788713;
          author: Joseph Lansdowne
"""

__kupfer_name__ = _("GNOME Session Management")
__kupfer_sources__ = ("GnomeItemsSource",)
__description__ = _("Special items and actions for GNOME environment")
__version__ = "2012-10-16"
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"
import typing as ty

from kupfer.plugin import session_support as support

if ty.TYPE_CHECKING:
    from gettext import gettext as _

# sequences of argument lists
LOGOUT_CMD = (
    ["gnome-panel-logout"],
    ["gnome-session-save", "--kill"],
    ["gnome-session-quit", "--logout"],
)
SHUTDOWN_CMD = (
    ["gnome-panel-logout", "--shutdown"],
    ["gnome-session-save", "--shutdown-dialog"],
    ["gnome-session-quit", "--power-off"],
)
LOCKSCREEN_CMD = (
    ["gnome-screensaver-command", "--lock"],
    ["xdg-screensaver", "lock"],
)


class GnomeItemsSource(support.CommonSource):
    source_scan_interval: int = 36000

    def __init__(self):
        support.CommonSource.__init__(self, _("GNOME Session Management"))

    def get_items(self):
        return (
            support.Logout(LOGOUT_CMD),
            support.LockScreen(LOCKSCREEN_CMD),
            support.Shutdown(SHUTDOWN_CMD),
        )



from __future__ import annotations

import typing as ty

# NOTE: TextRepresentation moved into obj.representation
# pylint: disable=unused-import
from kupfer.obj.representation import (
    TextRepresentation,
    UriListRepresentation,
)

if ty.TYPE_CHECKING:
    from gi.repository import Gtk

    from kupfer.obj.base import KupferObject

__all__ = (
    "TextRepresentation",
    "UriListRepresentation",
    "copy_to_clipboard",
    "get_text_representation",
)


def get_text_representation(obj: ty.Any) -> str | None:
    """Get text representation from any @obj."""
    try:
        return obj.get_text_representation()  # type: ignore
    except AttributeError:
        return None


def copy_to_clipboard(obj: KupferObject, clipboard: Gtk.Clipboard) -> bool:
    """Copy @obj to @clipboard, a Gtk.Clipboard.

    Support only text representation.
    TODO: support images (maybe)

    Return True if successful
    """

    try:
        text = obj.get_text_representation()  # type: ignore
    except AttributeError:
        pass
    else:
        clipboard.set_text(text, -1)  # -1 for computed string length
        return True

    return False

    # bellow part is legacy code that don't work anyway.


#     ## support copying text to clipboard
#     ## as well as files in both the general uri-list representation
#     ## and in nautilus' file copy clipboard type
#     (uri_id, text_id, nautilus_id) = (80, 81, 82)
#     nautilus_target = Gdk.Atom.intern("x-special/gnome-copied-files", False)

#     # udata is the data dict
#     # def store(clipboard, sdata, info, udata):
#     #     if info == uri_id:
#     #         sdata.set_uris(udata[uri_id])
#     #     if info == text_id:
#     #         sdata.set_text(udata[text_id])
#     #     if info == nautilus_id:
#     #         str_data_format = 8
#     #         sdata.set(nautilus_target, str_data_format, udata[nautilus_id])

#     # def clear(clipboard, udata):
#     #     pass

#     data: dict[int, ty.Any] = {}
#     try:
#         urilist = obj.get_urilist_representation()  # type: ignore
#     except AttributeError:
#         pass
#     else:
#         if urilist:
#             targets = Gtk.TargetList.new(None)
#             targets.add_uri_targets(uri_id)
#             # targets = Gtk.target_list_add_uri_targets(targets, uri_id)
#             targets.add(nautilus_target, 0, nautilus_id)
#             data[uri_id] = urilist
#             data[nautilus_id] = "copy\n" + "\n".join(urilist)

#     try:
#         text = obj.get_text_representation()  # type: ignore
#     except AttributeError:
#         pass
#     else:
#         targets = Gtk.TargetList.new(None)
#         targets.add_text_targets(text_id)
#         clipboard.set_text(text, -1)  # -1 for computed string length
#         data[text_id] = text

#     if data:
#         # F I X M E: How to set URIs on clipboard?
#         # clipboard.set_with_data(targets, store, clear, data)
#         # store all targets
#         # clipboard.set_can_store(targets)
#         return True

#     return False



__kupfer_name__ = _("Custom Terminal")
__description__ = _(
    """Configure a custom terminal emulator.

The plugin adds another terminal alternative (select it on the main page)."""
)
__version__ = ""
__author__ = "Ulrik Sverdrup"

import typing as ty

from kupfer import plugin_support, support
from kupfer.support import fileutils

if ty.TYPE_CHECKING:
    from gettext import gettext as _

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "command",
        "label": _("Command"),
        "type": str,
        "value": "",
    },
    {
        "key": "exearg",
        "label": _("Execute flag"),
        "type": str,
        "value": "-e",
        "tooltip": (
            "The flag which makes the terminal execute"
            " everything following it in the argument string. "
        ),
    },
)


def initialize_plugin(name):
    __kupfer_settings__.connect_settings_changed_cb(_update_alternative)
    _update_alternative()


def _update_alternative(*args):
    command = __kupfer_settings__["command"]
    exearg = __kupfer_settings__["exearg"]
    argv = support.argv_for_commandline(command)
    if not argv or not fileutils.lookup_exec_path(argv[0]):
        return

    plugin_support.register_alternative(
        __name__,
        "terminal",
        "custom1",
        name=_("Custom Terminal"),
        argv=argv,
        exearg=exearg,
        desktopid="",
        startup_notify=True,
    )



from __future__ import annotations

import typing as ty

from gi.repository import Gdk, Gtk

from kupfer import config, version

CheckCallback = ty.Callable[[str], bool]


class _GetKeyDialogController:
    """Get Key for binding dialog controller."""

    def __init__(
        self,
        check_callback: CheckCallback | None = None,
        previous_key: str | None = None,
        screen: Gdk.Screen | None = None,
        parent: Gtk.Window | None = None,
        show_clear: bool = True,
    ):
        """
        check_callback: optional function to check is entered key is valid.
        previous_key - optional previous keybinding, press equal act like cancel
        screen: Screen to use
        parent: Parent toplevel window
        show_clear: Show the “clear” button
        """
        builder = Gtk.Builder()
        builder.set_translation_domain(version.PACKAGE_NAME)

        builder.add_from_file(config.get_data_file("getkey_dialog.ui"))
        builder.connect_signals(self)  # pylint: disable=no-member
        self.window = builder.get_object("dialoggetkey")
        # self.labelkey = builder.get_object("labelkey")
        self.imagekeybindingaux = builder.get_object("imagekeybindingaux")
        self.labelkeybindingaux = builder.get_object("labelkeybindingaux")
        self.labelaccelerator = builder.get_object("labelaccelerator")
        if not show_clear:
            buttonclear = builder.get_object("buttonclear")
            buttonclear.hide()

        self.imagekeybindingaux.hide()
        self.labelkeybindingaux.hide()

        self._key: str | None = None
        self._check_callback = check_callback
        self._previous_key = previous_key
        self._press_time = None

        if screen:
            self.window.set_screen(screen)

        if parent:
            self.window.set_transient_for(parent)

        self.window.connect("focus-in-event", self._on_window_focus_in)
        self.window.connect("focus-out-event", self._on_window_focus_out)

    def run(self) -> str | None:
        """Run dialog, return key codes or None when user press cancel"""
        self.window.set_keep_above(True)
        self.window.run()
        self.window.destroy()
        return self._key

    def _return(self, key: str | None) -> None:
        "Finish dialog with @key as result"
        self._key = key
        self.window.hide()

    def on_buttoncancel_activate(self, _widget: Gtk.Widget) -> bool:
        self._return_cancel()
        return True

    def on_buttonclear_activate(self, _widget: Gtk.Widget) -> bool:
        self._return_clear()
        return True

    def _return_cancel(self) -> None:
        self._return(None)

    def _return_clear(self) -> None:
        self._return("")

    def _translate_keyboard_event(
        self, widget: Gtk.Widget, event: Gdk.EventKey
    ) -> tuple[int, int]:
        keymap = Gdk.Keymap.get_for_display(widget.get_display())
        # translate keys properly
        (
            _wasmapped,
            keyval,
            _egroup,
            _level,
            consumed,
        ) = keymap.translate_keyboard_state(
            event.hardware_keycode, event.get_state(), event.group
        )
        modifiers = Gtk.accelerator_get_default_mod_mask() & ~consumed
        state = event.get_state() & modifiers

        return keyval, state

    def _update_accelerator_label(self, keyval: int, state: int) -> None:
        accel_label = Gtk.accelerator_get_label(keyval, state)
        self.labelaccelerator.set_text(accel_label)

    def on_dialoggetkey_key_press_event(
        self, widget: Gtk.Widget, event: Gdk.EventKey
    ) -> None:
        self.imagekeybindingaux.hide()
        self.labelkeybindingaux.hide()
        self._press_time = event.time

        keyval, state = self._translate_keyboard_event(widget, event)
        state = Gdk.ModifierType(state)
        keyname = Gtk.accelerator_name(keyval, state)
        if keyname == "Escape":
            self._return_cancel()
        elif keyname == "BackSpace":
            self._return_clear()

        self._update_accelerator_label(keyval, state)

    def on_dialoggetkey_key_release_event(
        self, widget: Gtk.Widget, event: Gdk.EventKey
    ) -> None:
        if not self._press_time:
            return

        keyval, state = self._translate_keyboard_event(widget, event)
        self._update_accelerator_label(0, 0)

        state = Gdk.ModifierType(state)
        if Gtk.accelerator_valid(keyval, state):
            key = Gtk.accelerator_name(keyval, state)
            if self._previous_key is not None and key == self._previous_key:
                self._return_cancel()
                return

            if self._check_callback is None or self._check_callback(key):
                self._return(key)
            else:
                self.imagekeybindingaux.show()
                self.labelkeybindingaux.show()

    def _on_window_focus_in(
        self, _window: Gtk.Window, _event: Gdk.EventFocus
    ) -> None:
        pass

    def _on_window_focus_out(
        self, _window: Gtk.Window, _event: Gdk.EventFocus
    ) -> None:
        pass


def ask_for_key(
    check_callback: CheckCallback | None = None,
    previous_key: str | None = None,
    screen: Gdk.Screen | None = None,
    parent: Gtk.Window | None = None,
    show_clear: bool = True,
) -> str | None:
    dlg = _GetKeyDialogController(
        check_callback,
        previous_key,
        screen=screen,
        parent=parent,
        show_clear=show_clear,
    )
    return dlg.run()



__kupfer_name__ = _("Window List")
__kupfer_sources__ = ("WindowsSource", "WorkspacesSource")
__description__ = _("All windows on all workspaces")
__version__ = "2020-04-14"
__author__ = ""

import typing as ty

# pylint: disable=no-name-in-module
from gi.repository import Wnck

from kupfer.obj import Action, Leaf, Source
from kupfer.support import system, weaklib

if ty.TYPE_CHECKING:
    from gettext import gettext as _
    from gettext import ngettext


def _get_window(xid):
    """Get wnck window by xid."""
    if not xid:
        return None

    if scr := Wnck.Screen.get_default():
        for wnd in scr.get_windows():
            if wnd.get_xid() == xid:
                return wnd

    return None


def _get_workspace(idx):
    """Get wnck workspacke by its number."""
    if scr := Wnck.Screen.get_default():
        return scr.get_workspace(idx)

    return None


class WindowLeaf(Leaf):
    # object = window xid

    def get_actions(self):
        yield WindowActivateWorkspace()
        yield WindowMoveToWorkspace()
        yield WindowAction(_("Activate"), "activate", time=True)

        win = _get_window(self.object)
        if not win:
            return

        win_type = type(win)
        yield ToggleAction(
            _("Shade"),
            _("Unshade"),
            "shade",
            "unshade",
            win.is_shaded(),
            win_type.is_shaded,
        )
        yield ToggleAction(
            _("Minimize"),
            _("Unminimize"),
            "minimize",
            "unminimize",
            win.is_minimized(),
            win_type.is_minimized,
            time=True,
            icon="list-remove",
        )
        yield ToggleAction(
            _("Maximize"),
            _("Unmaximize"),
            "maximize",
            "unmaximize",
            win.is_maximized(),
            win_type.is_maximized,
            icon="list-add",
        )
        yield ToggleAction(
            _("Maximize Vertically"),
            _("Unmaximize Vertically"),
            "maximize_vertically",
            "unmaximize_vertically",
            win.is_maximized_vertically(),
            win_type.is_maximized_vertically,
            icon="list-add",
        )
        yield WindowAction(_("Close"), "close", time=True, icon="window-close")

    def is_valid(self):
        wnd = _get_window(self.object)
        return wnd and wnd.get_xid() == self.object

    def get_description(self):
        wnd = _get_window(self.object)
        if not wnd:
            return ""

        workspace = wnd.get_workspace()
        if not workspace:
            return ""

        _nr, name = workspace.get_number(), workspace.get_name()
        # TRANS: Window on (Workspace name), window description
        return _("Window on %(wkspc)s") % {"wkspc": name}

    def get_icon_name(self):
        return "kupfer-window"


class FrontmostWindow(WindowLeaf):
    qf_id = "frontwindow"

    def __init__(self):
        WindowLeaf.__init__(self, None, _("Frontmost Window"))

    # HACK: Make self.object a property
    # so that this leaf is *not* immutable
    def _set_object(self, obj):
        pass

    def _get_object(self):
        scr = Wnck.Screen.get_default()
        if scr is None:
            return None

        active = scr.get_active_window() or scr.get_previously_active_window()
        if (
            active
            and active.get_application().get_name()
            not in (
                "kupfer.py",
                system.get_application_filename(),
            )
            and not active.is_skip_tasklist()
        ):
            return active

        wspc = scr.get_active_workspace()
        for win in reversed(scr.get_windows_stacked()):
            if not win.is_skip_tasklist() and win.is_on_workspace(wspc):
                return win

        return None

    object = property(_get_object, _set_object)

    def repr_key(self):
        return None

    def get_description(self):
        return self.object and self.object.get_name()


class NextWindow(WindowLeaf):
    qf_id = "nextwindow"

    def __init__(self):
        WindowLeaf.__init__(self, None, _("Next Window"))

    def _set_object(self, obj):
        pass

    def _get_object(self):
        scr = Wnck.Screen.get_default()
        if scr is None:
            return None

        wspc = scr.get_active_workspace()
        for win in scr.get_windows_stacked():
            if not win.is_skip_tasklist() and win.is_on_workspace(wspc):
                return win

        return None

    object = property(_get_object, _set_object)

    def repr_key(self):
        return None

    def get_description(self):
        return self.object and self.object.get_name()


class WindowActivateWorkspace(Action):
    def __init__(self, name=_("Go To")):
        super().__init__(name)

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx

        window = _get_window(leaf.object)
        if not window:
            return

        workspace = window.get_workspace()
        time = ctx.environment.get_timestamp()
        workspace.activate(time)
        window.activate(time)

    def get_description(self):
        return _("Jump to this window's workspace and focus")

    def get_icon_name(self):
        return "go-jump"


class WindowMoveToWorkspace(Action):
    def __init__(self):
        Action.__init__(self, _("Move To..."))

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert iobj
        assert ctx

        window = _get_window(leaf.object)
        if not window:
            return

        workspace = _get_workspace(iobj.object)
        if not workspace:
            return

        window.move_to_workspace(workspace)
        time = ctx.environment.get_timestamp()
        workspace.activate(time)
        window.activate(time)

    def requires_object(self):
        return True

    def object_types(self):
        yield Workspace

    def object_source(self, for_item=None):
        return WorkspacesSource()

    def valid_object(self, iobj, for_item):
        window = _get_window(for_item.object)
        return window and window.get_workspace().get_number() != iobj.object

    def get_icon_name(self):
        return "forward"


class WindowAction(Action):
    def __init__(self, name, action, time=False, icon=None):
        super(Action, self).__init__(name)
        self.action = action or name.lower()
        self.time = time
        self.icon_name = icon

    def repr_key(self):
        return self.action

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx
        time = self._get_time(ctx) if self.time else None
        self._perform_action(self.action, leaf, time)

    @classmethod
    def _perform_action(cls, action_attr, leaf, time=None):
        window = _get_window(leaf.object)
        if not window:
            return

        action_method = getattr(window, action_attr)
        if time is not None:
            action_method(time)
        else:
            action_method()

    @classmethod
    def _get_time(cls, ctx):
        # @time will be != 0 if we are "inside"
        # a current gtk event
        return ctx.environment.get_timestamp()

    def get_icon_name(self):
        if not self.icon_name:
            return super().get_icon_name()

        return self.icon_name


class ToggleAction(WindowAction):
    """A toggle action, performing the enable / disable action as needed,
    for example minimize/unminimize.

    @istate: Initial state
    @predicate: Callable for state taking the window object as only argument
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        ename,
        uname,
        eaction,
        uaction,
        istate,
        predicate,
        time=False,
        icon=None,
    ):
        name = uname if istate else ename
        WindowAction.__init__(self, name, eaction, time=time, icon=icon)
        self.predicate = predicate
        self.uaction = uaction

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx

        wnd = _get_window(leaf.object)
        if not wnd:
            return

        if self.predicate(wnd):
            # only use time on the disable action
            time = self._get_time(ctx) if self.time else None
            self._perform_action(self.uaction, leaf, time)
        else:
            self._perform_action(self.action, leaf)


class WindowsSource(Source):
    def __init__(self, name=_("Window List")):
        super().__init__(name)

    def initialize(self):
        # "preload" windows: Ask for them early
        # since the first call "primes" the event loop
        # and always comes back empty
        if (screen := Wnck.Screen.get_default()) is not None:
            screen.get_windows_stacked()

    def is_dynamic(self):
        return True

    def get_items(self):
        # wnck should be "primed" now to return the true list
        screen = Wnck.Screen.get_default()
        if screen is None:
            self.output_debug("Environment not supported")
            return

        yield FrontmostWindow()
        yield NextWindow()

        for win in reversed(screen.get_windows_stacked()):
            if not win.is_skip_tasklist():
                name, app = (win.get_name(), win.get_application().get_name())
                if name != app and app not in name:
                    name = f"{name} ({app})"

                yield WindowLeaf(win.get_xid(), name)

    def get_description(self):
        return _("All windows on all workspaces")

    def get_icon_name(self):
        return "kupfer-window"

    def provides(self):
        yield WindowLeaf


class Workspace(Leaf):
    # object = number
    def get_actions(self):
        yield ActivateWorkspace()

    def repr_key(self):
        return self.object

    def get_icon_name(self):
        return "kupfer-window"

    def get_description(self):
        screen = Wnck.Screen.get_default()
        if screen:
            n_windows = sum(
                1
                for w in screen.get_windows()
                if (wsp := w.get_workspace())
                and wsp.get_number() == self.object
            )

            w_msg = ngettext("%d window", "%d windows", n_windows) % n_windows

            active_wspc = screen.get_active_workspace()
            if active_wspc.get_number() == self.object:
                return _("Active workspace") + " (" + w_msg + ")"

            if n_windows:
                return f"({w_msg})"

        return None


class ActivateWorkspace(Action):
    action_accelerator = "o"
    rank_adjust = 5

    def __init__(self):
        Action.__init__(self, _("Go To"))

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx

        if workspace := _get_workspace(leaf.object):
            time = ctx.environment.get_timestamp()
            workspace.activate(time)

    def get_description(self):
        return _("Jump to this workspace")

    def get_icon_name(self):
        return "go-jump"


class WorkspacesSource(Source):
    source_use_cache = False
    source_scan_interval: int = 3600

    def __init__(self):
        super().__init__(_("Workspaces"))

    def initialize(self):
        if (screen := Wnck.Screen.get_default()) is not None:
            screen.get_workspaces()
            weaklib.gobject_connect_weakly(
                screen, "workspace-created", self._changed
            )
            weaklib.gobject_connect_weakly(
                screen, "workspace-destroyed", self._changed
            )

    def _changed(self, screen, workspace):
        self.mark_for_update()

    def get_items(self):
        # wnck should be "primed" now to return the true list
        screen = Wnck.Screen.get_default()
        if screen is None:
            return

        for wspc in screen.get_workspaces():
            yield Workspace(wspc.get_number(), wspc.get_name())

    def get_icon_name(self):
        return "kupfer-window"

    def provides(self):
        yield Workspace



import typing as ty

if ty.TYPE_CHECKING:
    from gettext import gettext as _

VERSION = "development version"
PACKAGE_NAME = "kupfer"

try:
    from kupfer import version_subst  # type:ignore
except ImportError:
    pass
else:
    VERSION = version_subst.VERSION
    PACKAGE_NAME = version_subst.PACKAGE_NAME

ICON_NAME = "kupfer"
DESKTOP_ID = "kupfer"
PROGRAM_NAME = _("Kupfer")

AUTHORS = """Ulrik Sverdrup
Karol Będkowski
Francesco Marella
Chmouel Boudjnah
Horia V. Corcalciuc
Grigory Javadyan
Chris Parsons
Fabian Carlström
Jakh Daven
Thomas Renard
""".splitlines()

PACKAGERS = """
Luca Falavigna (Debian, Ubuntu)
Andrew from WebUpd8 (Ubuntu PPA ca 2017)
Francesco Marella (Ubuntu PPA ca 2010)
D. Can Celasun (Arch Linux AUR)
""".splitlines()

TRANSLATORS = """
Marek Černocký (cs)
Petr Kovar (cs)
Joe Hansen (da)
Thibaud Roth (de)
Mario Blättermann (de)
Leandro Leites (es)
Jesús Barbero Rodríguez (es)
Jorge González (es)
Daniel Mustieles (es)
Oier Mees (eu)
Iñaki Larrañaga Murgoitio (eu)
Christophe Benz (fr)
Marcos Lans (gl)
Fran Diéguez (gl)
Andrea Zagli (it)
Francesco Marella (it)
Martin Koelewijn (nl)
Kjartan Maraas (no)
Maciej Kwiatkowski (pl)
Karol Będkowski (pl)
Carlos Pais (pt)
Andrej Žnidaršič (sl)
Matej Urbančič (sl)
M. Deran Delice (tr)
lh (zh_CN)
Aron Xu (zh_CN)
Yinghua Wang (zh_CN)
""".splitlines()

ARTISTS = """
Nasser Alshammari <designernasser@gmail.com> (Kupfer Icon)
GNOME Project https://www.gnome.org (Misc Icons)
""".splitlines()

AUTHORS += ARTISTS + PACKAGERS + TRANSLATORS

DOCUMENTERS: list[str] = []

# TRANS: Don't translate literally!
# TRANS: This should be a list of all translators of this language
TRANSLATOR_CREDITS = _("translator-credits")

WEBSITE = "https://kupferlauncher.github.io/"
HELP_WEBSITE = "https://kupferlauncher.github.io/help/"

SHORT_DESCRIPTION = _("A free software (GPLv3+) launcher")
COPYRIGHT = """Copyright © 2007–2023 Ulrik Sverdrup with others"""

LICENSE = _(
    """
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
)

# follows strings used elsewhere

_("Could not find running Kupfer")



__kupfer_name__ = _("Send Keys")
__kupfer_actions__ = ("CopyAndPaste", "SendKeys", "TypeText")
__description__ = _("Send synthetic keyboard events using xautomation")
__version__ = ""
__author__ = ""

import typing as ty
from contextlib import suppress

from gi.repository import Gdk, Gtk

from kupfer import interface, launch
from kupfer.obj import Action, Leaf, OperationError, TextLeaf

if ty.TYPE_CHECKING:
    from gettext import gettext as _

# delay for first keypress and all following
_INIT_DELAY = "usleep 300000"
_INTER_DELAY = "usleep 50000"


class CopyAndPaste(Action):
    # rank down since it applies everywhere
    rank_adjust = -2

    def __init__(self):
        Action.__init__(self, _("Paste to Foreground Window"))

    def activate(self, leaf, iobj=None, ctx=None):
        clip = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        interface.copy_to_clipboard(leaf, clip)
        xte_paste_argv = [
            "xte",
            _INIT_DELAY,
            "keydown Control_L",
            "key v",
            "keyup Control_L",
        ]
        try:
            launch.spawn_async_raise(xte_paste_argv)
        except launch.SpawnError as exc:
            raise OperationError(exc) from exc

    def item_types(self):
        yield Leaf

    def valid_for_item(self, leaf):
        with suppress(AttributeError):
            return bool(interface.get_text_representation(leaf))

    def get_description(self):
        return _("Copy to clipboard and send Ctrl+V to foreground window")

    def get_icon_name(self):
        return "edit-paste"


class SendKeys(Action):
    def __init__(self):
        Action.__init__(self, _("Send Keys"))

    def activate(self, leaf, iobj=None, ctx=None):
        return self.activate_multiple((leaf,))

    def activate_multiple(self, objects):
        xte_sendkey_argv = ["xte", _INIT_DELAY]
        iterobjects = iter(objects)
        for obj in iterobjects:
            xte_sendkey_argv.extend(self.make_keystr_arguments(obj.object))
            break

        for obj in iterobjects:
            xte_sendkey_argv.append(_INTER_DELAY)
            xte_sendkey_argv.extend(self.make_keystr_arguments(obj.object))

        try:
            launch.spawn_async_raise(xte_sendkey_argv)
        except launch.SpawnError as exc:
            raise OperationError(exc) from exc

    def make_keystr_arguments(self, keystr):
        keys, orig_mods = Gtk.accelerator_parse(keystr)
        modifiers = {
            Gdk.ModifierType.SHIFT_MASK: "Shift_L",
            Gdk.ModifierType.CONTROL_MASK: "Control_L",
            Gdk.ModifierType.SUPER_MASK: "Super_L",
            Gdk.ModifierType.MOD1_MASK: "Alt_L",  # pylint: disable=no-member
        }
        mod_names = []
        mods = orig_mods
        for mod, value in modifiers.items():
            if mod & mods:
                mod_names.append(value)
                mods &= ~mod

        if mods != 0:
            label = Gtk.accelerator_get_label(keys, orig_mods)
            raise OperationError(f"Keys not yet implemented: {label}")

        mods_down = ["keydown " + n for n in mod_names]
        mods_up = ["keyup " + n for n in reversed(mod_names)]
        return [*mods_down, f"key {Gdk.keyval_name(keys)}", *mods_up]

    def item_types(self):
        yield TextLeaf

    def valid_for_item(self, leaf):
        text = leaf.object
        keys, _mods = Gtk.accelerator_parse(text)
        return bool(keys)

    def get_description(self):
        return _("Send keys to foreground window")


class TypeText(Action):
    rank_adjust = -2

    def __init__(self):
        Action.__init__(self, _("Type Text"))

    def activate(self, leaf, iobj=None, ctx=None):
        text = interface.get_text_representation(leaf)
        if not text:
            return

        xte_paste_argv = ["xte", "usleep 300000"]
        # replace all newlines with 'key Return'
        for line in text.splitlines(True):
            xte_paste_argv.append("str " + line.rstrip("\r\n"))
            if line.endswith("\n"):
                xte_paste_argv.append("key Return")

        try:
            launch.spawn_async_raise(xte_paste_argv)
        except launch.SpawnError as exc:
            raise OperationError(exc) from exc

    def item_types(self):
        yield Leaf

    def valid_for_item(self, leaf):
        with suppress(AttributeError):
            return bool(interface.get_text_representation(leaf))

    def get_description(self):
        return _("Type the text to foreground window")



from __future__ import annotations

__kupfer_name__ = _("Favorites")
__kupfer_sources__ = ("FavoritesSource",)
__kupfer_actions__ = ("AddFavorite", "RemoveFavorite")
__description__ = _("Mark commonly used items and store objects for later use")
__version__ = "2009-12-30"
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import typing as ty
import weakref

from kupfer import puid

# NOTE: core import
from kupfer.core import learn
from kupfer.obj import Action, Leaf, Source, SourceLeaf

if ty.TYPE_CHECKING:
    from gettext import gettext as _


class FavoritesSource(Source):
    """Keep a list of Leaves that the User may add and remove from"""

    instance: FavoritesSource = None  # type:ignore
    source_scan_interval: int = 3600

    def __init__(self):
        Source.__init__(self, _("Favorites"))
        ## these are default favorites for new users
        self.references: list[puid.PuID] = [
            "<kupfer.plugin.core.contents.Help>",
            "<kupfer.plugin.core.contents.Preferences>",
        ]

    def config_save(self) -> dict[str, ty.Any]:
        references = [puid.get_unique_id(F) for F in self.favorites]
        return {"favorites": references, "version": self.version}

    def config_save_name(self) -> str:
        return __name__

    def config_restore(self, state: dict[str, ty.Any]) -> None:
        self.references = state["favorites"]

    def _find_item(self, id_: puid.PuID) -> Leaf | None:
        itm = puid.resolve_unique_id(id_, excluding=self)
        if itm is None:
            return None

        assert isinstance(itm, Leaf)

        # ignore invalid objects
        if hasattr(itm, "is_valid") and not itm.is_valid():
            return None

        if puid.is_reference(id_):
            assert isinstance(id_, str)
            self.reference_table[id_] = itm
        else:
            assert isinstance(id_, puid.SerializedObject)
            self.persist_table[id_] = itm

        return itm

    # pylint: disable=attribute-defined-outside-init
    def initialize(self):
        FavoritesSource.instance = self
        self.favorites: list[Leaf] = []
        # persist_table map Serialized object to Leaves
        self.persist_table: dict[puid.SerializedObject, Leaf] = {}
        # reference table map reference-type id to leaf
        self.reference_table: weakref.WeakValueDictionary[str, Leaf] = (
            weakref.WeakValueDictionary()
        )
        self.mark_for_update()

    def finalize(self) -> None:
        learn.replace_favorites(__name__)

    def _update_items(self):
        # insert items on beginning to make last added items first on list
        favorites: list[Leaf] = []
        for id_ in self.references:
            if isinstance(id_, puid.SerializedObject) and (
                leaf := self.persist_table.get(id_)
            ):
                # id_ is in persist_table so is SerializedObject
                favorites.insert(0, leaf)
                continue

            if isinstance(id_, str) and (
                leaf := self.reference_table.get(id_)
            ):
                # id_ is in reference_table, so is reference, so it's str
                favorites.insert(0, leaf)
                continue

            if (itm := self._find_item(id_)) is not None:
                favorites.insert(0, itm)
            else:
                self.output_debug("MISSING:", id_)

        self.favorites = favorites

    def add(self, itm):
        if id_ := puid.get_unique_id(itm):
            # there shouldn't be possible to add twice the same leaf so
            # remove should be not needed.
            self.remove(itm)
            learn.add_favorite(__name__, itm)
            # favorites will be rebuild, so we can append
            self.favorites.append(itm)
            self.references.append(id_)
            self.mark_for_update()

    def has_item(self, itm: Leaf) -> bool:
        return itm in self.favorites

    def remove(self, itm: Leaf) -> None:
        if itm not in self.favorites:
            return

        learn.remove_favorite(__name__, itm)
        self.favorites.remove(itm)
        if id_ := puid.get_unique_id(itm):
            try:
                self.references.remove(id_)
            except KeyError:
                for key, val in self.persist_table.items():
                    if val == itm:
                        self.references.remove(key)
                        self.persist_table.pop(key)
                        break

        self.mark_for_update()

    def get_items(self):
        self._update_items()
        learn.replace_favorites(__name__, *self.favorites)

        return self.favorites

    def get_description(self):
        return _('Shelf of "Favorite" items')

    def get_icon_name(self):
        return "emblem-favorite"

    def provides(self):
        # returning nothing means it provides anything
        return ()


class AddFavorite(Action):
    # Rank down, since it applies everywhere
    rank_adjust = -15

    def __init__(self):
        Action.__init__(self, _("Add to Favorites"))

    def activate(self, leaf, iobj=None, ctx=None):
        FavoritesSource.instance.add(leaf)

    def item_types(self):
        yield Leaf

    def valid_for_item(self, leaf):
        return not isinstance(
            leaf, SourceLeaf
        ) and not FavoritesSource.instance.has_item(leaf)

    def get_description(self):
        return _("Add item to favorites shelf")

    def get_icon_name(self):
        return "list-add"


class RemoveFavorite(Action):
    rank_adjust = -15

    def __init__(self):
        Action.__init__(self, _("Remove from Favorites"))

    def activate(self, leaf, iobj=None, ctx=None):
        FavoritesSource.instance.remove(leaf)

    def item_types(self):
        yield Leaf

    def valid_for_item(self, leaf):
        return FavoritesSource.instance.has_item(leaf)

    def get_description(self):
        return _("Remove item from favorites shelf")

    def get_icon_name(self):
        return "list-remove"



import typing as ty

from gi.repository import Gio

from kupfer import icons, plugin_support
from kupfer.support import desktop_parse

if ty.TYPE_CHECKING:
    from gettext import gettext as _


def initialize_alternatives(name):
    plugin_support.register_alternative(
        name,
        "icon_renderer",
        "gtk",
        name=_("GTK+"),
        renderer=icons.IconRenderer,
    )

    plugin_support.register_alternative(
        name,
        "terminal",
        "gnome-terminal",
        name=_("GNOME Terminal"),
        argv=["gnome-terminal"],
        exearg="-x",
        desktopid="gnome-terminal.desktop",
        startup_notify=True,
    )

    plugin_support.register_alternative(
        name,
        "terminal",
        "xfce4-terminal",
        name=_("XFCE Terminal"),
        argv=["xfce4-terminal"],
        exearg="-x",
        desktopid="xfce4-terminal.desktop",
        startup_notify=True,
    )

    plugin_support.register_alternative(
        name,
        "terminal",
        "exo-open",
        name="exo-open",
        argv=["exo-open", "--launch", "TerminalEmulator"],
        exearg="",
        desktopid="",
        startup_notify=False,
    )

    plugin_support.register_alternative(
        name,
        "terminal",
        "lxterminal",
        name=_("LXTerminal"),
        argv=["lxterminal"],
        exearg="-e",
        desktopid="lxterminal.desktop",
        startup_notify=False,
    )

    plugin_support.register_alternative(
        name,
        "terminal",
        "xterm",
        name=_("X Terminal"),
        argv=["xterm"],
        exearg="-e",
        desktopid="xterm.desktop",
        startup_notify=False,
    )

    plugin_support.register_alternative(
        name,
        "terminal",
        "x-terminal-emulator",
        name="x-terminal-emulator",
        argv=["x-terminal-emulator"],
        exearg="-e",
        desktopid="",
        startup_notify=False,
    )

    plugin_support.register_alternative(
        name,
        "terminal",
        "urxvt",
        name=_("Urxvt"),
        argv=["urxvt"],
        exearg="-e",
        desktopid="urxvt.desktop",
        startup_notify=False,
    )

    plugin_support.register_alternative(
        name,
        "terminal",
        "konsole",
        name=_("Konsole"),
        argv=["konsole"],
        exearg="-e",
        desktopid="konsole.desktop",
        # Not sure here, so setting to false
        startup_notify=False,
    )

    _register_terminal_emulators(name)
    _register_text_editors(name)


def _register_text_editors(name):
    plugin_support.register_alternative(
        name,
        "editor",
        "sys-editor",
        name=_("System `editor`"),
        argv=["editor"],
        terminal=True,
    )

    # find all applications that support text/plain and register it
    # as text editors

    for app in Gio.app_info_get_all_for_type("text/plain"):
        app_id = app.get_id()
        cmd = app.get_commandline()
        args = desktop_parse.parse_argv(cmd)

        plugin_support.register_alternative(
            name,
            "editor",
            app_id.removesuffix(".desktop"),
            name=app.get_display_name(),
            argv=args,
            terminal=app.get_boolean("Terminal"),
        )


def _register_terminal_emulators(name):
    """Find all applications with category 'TerminalEmulator' and register
    it as terminal alternative."""

    # hard coded skip-list for already defined above terminals but with
    # different id
    skip = ("debian-uxterm.desktop", "debian-xterm.desktop")

    for app in Gio.app_info_get_all():
        app_id = app.get_id()
        if app_id in skip:
            continue

        if "TerminalEmulator" not in (app.get_categories() or ""):
            continue

        cmd = app.get_commandline()
        args = desktop_parse.parse_argv(cmd)

        plugin_support.register_alternative(
            name,
            "terminal",
            app_id.removesuffix(".desktop"),
            name=app.get_display_name(),
            argv=args,
            exearg="",
            desktopid=app_id,
            startup_notify=False,
        )



from __future__ import annotations

__kupfer_name__ = _("URL Actions")
__kupfer_sources__ = ()
__kupfer_text_sources__ = ()
__kupfer_actions__ = ("DownloadAndOpen", "DownloadTo")
__description__ = _("URL Actions")
__version__ = ""
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import os
import shutil
import typing as ty
import urllib.error
import urllib.parse
import urllib.request

from kupfer import launch
from kupfer.obj import Action, FileLeaf, UrlLeaf
from kupfer.support import fileutils, task

if ty.TYPE_CHECKING:
    import http.client
    from gettext import gettext as _


def get_dest_name(response: http.client.HTTPResponse) -> str:
    headers = response.headers

    # try get filename from content-disposition header
    content_disp = headers.get("Content-Disposition", "")
    for part in content_disp.split(";"):
        key, _sep, val = part.strip().partition("=")
        if key.lower() == "filename":
            return val

    name = "index"
    # try get filename from url
    url = urllib.parse.urlparse(response.url)
    if url.path not in ("", "/"):
        name = os.path.basename(url.path.rstrip("/"))
        # if name have extension - return it
        if os.path.splitext(name)[1]:
            return name

    # if name not contain extension, try guess it from content-type.
    # only basic types
    content_type = headers.get("Content-Type", "")
    content_type = content_type.split(";", 1)[0].strip()
    if content_type == "text/html":
        return f"{name}.html"

    if content_type == "text/plain":
        return f"{name}.txt"

    return name


class DownloadTask(task.ThreadTask):
    def __init__(
        self, uri, destdir=None, tempfile=False, finish_callback=None
    ):
        super().__init__()
        self.uri = uri
        self.download_finish_callback = finish_callback
        self.destdir = destdir
        self.use_tempfile = tempfile
        self.destpath = None

    def _get_dst_file(
        self, destname: str
    ) -> tuple[ty.BinaryIO | None, str | None]:
        if self.use_tempfile:
            return fileutils.get_safe_tempfile()

        return fileutils.get_destfile_in_directory(self.destdir, destname)

    def thread_do(self):
        with urllib.request.urlopen(self.uri) as response:
            if response.status >= 300 or response.status < 200:  # noqa: PLR2004
                raise RuntimeError(
                    f"Could not download file; status: {response.status}"
                )

            destname = get_dest_name(response)
            destfile, self.destpath = self._get_dst_file(destname)
            if not destfile:
                raise OSError("Could not write output file")

            try:
                shutil.copyfileobj(response, destfile)
            finally:
                destfile.close()

    def thread_finish(self):
        if self.download_finish_callback:
            self.download_finish_callback(self.destpath)


class DownloadAndOpen(Action):
    """Asynchronous action to download file and open it"""

    def __init__(self):
        Action.__init__(self, _("Download and Open"))

    def is_async(self):
        return True

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx

        def finish_action(filename):
            launch.show_path(filename)
            ctx.register_late_result(FileLeaf(filename), show=False)

        uri = leaf.object
        return DownloadTask(uri, None, True, finish_action)

    def item_types(self):
        yield UrlLeaf

    def get_description(self):
        return None


class DownloadTo(Action):
    def __init__(self):
        Action.__init__(self, _("Download To..."))

    def is_async(self):
        return True

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx
        assert iobj

        def finish_action(filename):
            ctx.register_late_result(FileLeaf(filename))

        uri = leaf.object
        return DownloadTask(uri, iobj.object, False, finish_action)

    def item_types(self):
        yield UrlLeaf

    def requires_object(self):
        return True

    def object_types(self):
        yield FileLeaf

    def valid_object(self, obj, for_item=None):
        return fileutils.is_directory_writable(obj.object)

    def get_description(self):
        return _("Download URL to a chosen location")



"""
This file import most common objects, so can they can be imported
directly from kupfer.obj
"""

from kupfer.obj import fileactions
from kupfer.obj.actions import Execute, OpenTerminal, OpenUrl, Perform
from kupfer.obj.apps import AppLeaf
from kupfer.obj.base import (
    Action,
    AnySource,
    KupferObject,
    Leaf,
    Source,
    TextSource,
)
from kupfer.obj.exceptions import NotAvailableError, OperationError
from kupfer.obj.files import FileLeaf
from kupfer.obj.objects import RunnableLeaf, SourceLeaf, TextLeaf, UrlLeaf

# importint fileactions here prevent circular imports
__dummy = fileactions

__all__ = (
    "Action",
    "AnySource",
    "AppLeaf",
    "Execute",
    "FileLeaf",
    "FileLeaf",
    "KupferObject",
    "Leaf",
    "NotAvailableError",
    "OpenTerminal",
    "OpenUrl",
    "OperationError",
    "Perform",
    "RunnableLeaf",
    "Source",
    "SourceLeaf",
    "TextLeaf",
    "TextSource",
    "UrlLeaf",
)



from __future__ import annotations

import sys
import traceback
import typing as ty
from time import time as timestamp

if ty.TYPE_CHECKING:
    from kupfer.support.types import ExecInfo

__all__ = (
    "OutputMixin",
    "print_debug",
    "print_error",
    "print_exc",
    "print_info",
    "timing_start",
    "timing_step",
)

DEBUG = False
COLORS = True

_COLOR_INFO = "\033[96m"
_COLOR_WARNING = "\033[93m"
_COLOR_FAIL = "\033[91m"
_COLOR_STD = "\033[0m"


class OutputMixin:
    """A mixin class providing prefixed output standard output and DEBUG output."""

    def _output_category(self) -> str:
        return f"[{type(self).__module__}] {type(self).__name__}:"

    def _output_core(
        self,
        prefix: str,
        sep: str,
        end: str,
        stream: ty.TextIO | None,
        *items: ty.Any,
    ) -> None:
        category = self._output_category()
        pitems = [
            item if isinstance(item, (str, int, float)) else repr(item)
            for item in items
        ]
        print(f"{prefix}{category}", *pitems, sep=sep, end=end, file=stream)

    def output_info(
        self, *items: ty.Any, sep: str = " ", end: str = "\n", **kwargs: ty.Any
    ) -> None:
        """Output given items using @sep as separator, ending the line with @end"""
        if COLORS:
            self._output_core(
                f"{_COLOR_INFO}INF{_COLOR_STD} ", sep, end, sys.stdout, *items
            )
        else:
            self._output_core("INF ", sep, end, sys.stdout, *items)

    def output_exc(self, exc_info: ExecInfo | None = None) -> None:
        """Output current exception, or use @exc_info if given"""
        etype, value, tback = exc_info or sys.exc_info()
        assert etype
        if DEBUG:
            if COLORS:
                self._output_core(
                    f"{_COLOR_FAIL}EXC{_COLOR_STD} ",
                    "",
                    "\n",
                    sys.stderr,
                )
            else:
                self._output_core("EXC ", "", "\n", sys.stderr)

            traceback.print_exception(etype, value, tback, file=sys.stderr)

            return

        msg = f"{etype.__name__}: {value}"
        if COLORS:
            self._output_core(
                f"{_COLOR_FAIL}EXC{_COLOR_STD} ",
                " ",
                "\n",
                sys.stderr,
                msg,
            )
        else:
            self._output_core("EXC ", " ", "\n", sys.stderr, msg)

    def output_debug(
        self, *items: ty.Any, sep: str = " ", end: str = "\n", **kwargs: ty.Any
    ) -> None:
        if DEBUG:
            self._output_core("DBG ", sep, end, sys.stderr, *items)

    def output_error(
        self, *items: ty.Any, sep: str = " ", end: str = "\n", **kwargs: ty.Any
    ) -> None:
        if COLORS:
            self._output_core(
                f"{_COLOR_WARNING}ERR{_COLOR_STD} ",
                sep,
                end,
                sys.stderr,
                *items,
            )
        else:
            self._output_core("ERR ", sep, end, sys.stderr, *items)


class _StaticOutput(OutputMixin):
    current_calling_module: str | None = None

    def _output_category(self) -> str:
        return f"[{self.current_calling_module}]:"

    def print_info(
        self, modulename: str, *args: ty.Any, **kwargs: ty.Any
    ) -> None:
        self.current_calling_module = modulename
        self.output_info(*args, **kwargs)

    def print_error(
        self, modulename: str, *args: ty.Any, **kwargs: ty.Any
    ) -> None:
        self.current_calling_module = modulename
        self.output_error(*args, **kwargs)

    def print_exc(
        self, modulename: str, *args: ty.Any, **kwargs: ty.Any
    ) -> None:
        self.current_calling_module = modulename
        self.output_exc(*args, **kwargs)

    def print_debug(
        self, modulename: str, *args: ty.Any, **kwargs: ty.Any
    ) -> None:
        if DEBUG:
            self.current_calling_module = modulename
            self.output_debug(*args, **kwargs)


_StaticOutputInst = _StaticOutput()

print_info = _StaticOutputInst.print_info
print_debug = _StaticOutputInst.print_debug
print_error = _StaticOutputInst.print_error
print_exc = _StaticOutputInst.print_exc


def timing_start() -> float | None:
    """In debug get current timestamp; return None otherwise"""
    if DEBUG:
        return timestamp()

    return None


def timing_step(
    modulename: str, start: float | None, label: str, *args: str
) -> float | None:
    """In debug printtime from `start`."""
    if DEBUG and start:
        cts = timestamp()
        print_debug(modulename, label, f"in {cts - start:.6f} s", *args)
        return cts

    return None



from __future__ import annotations

__kupfer_name__ = _("Textutils")
__kupfer_actions__ = (
    "Convert",
    "LineConvert",
    "Format",
)
__kupfer_sources__ = ("Generators",)
__description__ = _("Action for text, useful text generators and converters.")
__version__ = "2023.1"
__author__ = "Karol Będkowski"

import base64
import datetime
import json
import secrets
import string
import typing as ty
import urllib.parse
import uuid
import xml.dom.minidom
from contextlib import suppress
from functools import partial

from kupfer import icons
from kupfer.obj import (
    Action,
    Leaf,
    OperationError,
    RunnableLeaf,
    Source,
    TextLeaf,
)
from kupfer.ui import getdata_dialog

if ty.TYPE_CHECKING:
    from gettext import gettext as _

TextConverter = ty.Callable[[str], ty.Optional[str]]


class _Converter(Leaf):
    object: TextConverter

    def __init__(
        self, func: TextConverter, name: str, description: str | None = None
    ):
        super().__init__(func, name)
        self.description = description

    def get_description(self) -> str | None:
        return self.description


class _ConvertersSource(Source):
    def __init__(
        self,
        items_gen: ty.Callable[[Leaf], ty.Iterator[_Converter]],
        for_item: Leaf,
    ) -> None:
        super().__init__("Converters")
        self.items_gen = items_gen
        self.for_item = for_item

    def get_items(self):
        yield from self.items_gen(self.for_item)


class _ConvertAction(Action):
    def has_result(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert isinstance(leaf, TextLeaf)
        assert iobj and isinstance(iobj, _Converter)

        converter = iobj.object
        try:
            if (result := converter(leaf.object)) is not None:
                return TextLeaf(result)
        except Exception as err:
            raise OperationError(f"Error: {err}") from err

        return None

    def requires_object(self):
        return True

    def object_source(self, for_item=None):
        return _ConvertersSource(self._get_converters, for_item)

    def object_types(self):
        yield _Converter

    def item_types(self):
        yield TextLeaf

    def _get_converters(self, for_leaf: Leaf) -> ty.Iterator[_Converter]:
        raise NotImplementedError

    def get_gicon(self):
        return icons.ComposedIcon("text-x-generic", "kupfer-execute")

    def get_icon_name(self):
        return "text-x-generic"


def _str_to_unix_ts(inp: str) -> str:
    """Convert int or float from `inp` do date."""
    ts = float(inp)
    # check is ns/ms precision
    if ts > 9999999999999.0:  # noqa: PLR2004
        ts /= 1000000.0
    elif ts > 9999999999.0:  # noqa: PLR2004
        ts /= 1000.0

    return str(datetime.datetime.fromtimestamp(ts))


def _uncamellcase(text: str) -> str:
    """Convert 'SomeTextInCamelCase' into 'some text in camel case'."""
    res: list[str] = []
    for let in text:
        if let.isupper():
            if res and res[-1] != " ":
                res.append(" ")

            res.append(let.lower())

        else:
            res.append(let)

    return "".join(res)


def _camelcase(instr: str) -> str:
    words = instr.split(" ")
    if words:
        return words[0].lower() + "".join(p.capitalize() for p in words[1:])

    return instr


def _trim_to_len_sep(
    instr: str, max_len: int, sep: str, min_len: int = 10
) -> str:
    """If length of `instr` excess `max_len` trim it to this length.
    If `sep` is given also try to trim string to last found `sep` but
    keep `min_len` characters."""
    if len(instr) > max_len:
        instr = instr[:max_len]
        if sep and (idx := instr.rfind(sep)) > min_len:
            instr = instr[:idx]

    return instr


def _to_filename(instr: str) -> str:
    """Remove/replace characters that may be not allowed in filename.
    May be not necessary for posix filesystems, but there are still
    fat/ntfs/etc around."""
    unprintable = (
        "\x00\x01\x02\x03\x04\x05\x06\x07\x08\x0e\x0f"
        "\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a"
        "\x1b\x1c\x1d\x1e\x1f\x7f"
    )
    transtable = str.maketrans(
        "\n\r/\\ :<>?*|\"'", "__--_-_______", unprintable
    )
    res = (
        instr.strip()
        .translate(transtable)
        .replace("-_", "-")
        .replace("_-", "-")
    )
    while "--" in res:
        res = res.replace("__", "_")

    if len(res) > 255:  # noqa: PLR2004
        base, _sep, ext = res.rpartition(".")
        if ext and len(ext) < len(base):
            base = _trim_to_len_sep(base, 250 - len(ext), "_")
            return f"{base}.{ext}"

    return _trim_to_len_sep(res, 255, "_")


class Convert(_ConvertAction):
    def __init__(self, name=_("Convert…")):
        super().__init__(name)

    def get_description(self):
        return _("Convert text with selected tool")

    def _get_converters(self, for_leaf: Leaf) -> ty.Iterator[_Converter]:
        yield _Converter(
            lambda x: base64.b64encode(x.encode()).decode(),
            _("Encode text with Base64"),
        )
        yield _Converter(
            lambda x: base64.b64decode(x.encode()).decode(),
            _("Decode text with Base64"),
        )
        yield _Converter(
            urllib.parse.quote,
            _("Quote URL"),
            _("Replace special characters in string using the %%xx escape."),
        )
        yield _Converter(
            urllib.parse.unquote,
            _("Unquote URL"),
            _("Replace %%xx escapes with their single-character equivalent."),
        )

        with suppress(ValueError):
            if float(for_leaf.object) > 0.0:
                yield _Converter(_str_to_unix_ts, _("Unix timestamp to date"))

        yield _Converter(str.lower, _("to lower case"))
        yield _Converter(str.upper, _("to upper case"))
        yield _Converter(str.capitalize, _("to sentence case"))
        yield _Converter(
            lambda x: " ".join(p.capitalize() for p in x.split(" ")),
            _("Capitalize words"),
            _("Convert 'some string' to 'Some String'"),
        )
        yield _Converter(
            _camelcase,
            _("To camel case"),
            _("Convert 'Some string' to 'comeString'"),
        )
        yield _Converter(
            lambda x: "".join(p.capitalize() for p in x.split(" ")),
            _("To pascal case"),
            _("Convert 'Some string' to 'SomeString'"),
        )
        yield _Converter(
            lambda x: x.upper().replace(" ", "_"),
            _("To constant case"),
            _("Convert 'some string' to 'SOME_STRING'"),
        )
        yield _Converter(
            _uncamellcase,
            _("Camel case to lowercase"),
            _("Convert 'SomeString' do 'some string'"),
        )
        yield _Converter(
            lambda x: x.lower().replace("_", " "),
            _("Constant case to lowercase"),
            _("Convert 'SOME_STRING' do 'some string'"),
        )
        yield _Converter(
            _to_filename,
            _("To valid filename"),
            _("Convert string to usable file name"),
        )


def _ask_and_join_lines(inp: str) -> str | None:
    res = getdata_dialog.ask_for_text(
        _("Join lines options"),
        _("Please enter characters to join with"),
        _("Separator"),
        ";",
    )

    if res:
        return res.join(filter(None, map(str.strip, inp.split("\n"))))

    return None


def _ask_and_quote_join_lines(inp: str) -> str | None:
    dlg = getdata_dialog.GetDataDialog(_("Quote and join lines options"))

    dlg.add_field("sep", _("Separator:"), ";")
    dlg.add_field("qchar", _("Quotation character:"), '"')
    res = dlg.run()
    if res and (sep := res["sep"]) and (quot := res["qchar"]):
        return sep.join(
            f"{quot}{t}{quot}" for i in inp.split("\n") if (t := i.strip())
        )

    return None


class LineConvert(_ConvertAction):
    def __init__(self, name=_("Convert lines…")):
        super().__init__(name)

    def get_description(self):
        return _("Convert multiline text with selected tool")

    def _get_converters(self, for_leaf: Leaf) -> ty.Iterator[_Converter]:
        yield _Converter(
            _ask_and_join_lines,
            _("Join lines with..."),
            _("Join lines with user selected separator"),
        )
        yield _Converter(
            _ask_and_quote_join_lines,
            _("Quote and join lines..."),
            _("Wrap with quote characters and join lines."),
        )


class Format(_ConvertAction):
    def __init__(self, name=_("Format…")):
        super().__init__(name)

    def get_description(self):
        return _("Format text using selected formatter")

    def _get_converters(self, for_leaf: Leaf) -> ty.Iterator[_Converter]:
        yield _Converter(
            lambda x: "\n".join(line.strip() for line in x.split("\n")),
            _("strip whitespaces"),
        )
        yield _Converter(
            lambda x: json.dumps(json.loads(x), sort_keys=True, indent=4),
            _("format json"),
        )
        yield _Converter(
            lambda x: json.dumps(json.loads(x), separators=(",", ":")),
            _("compress json"),
        )
        yield _Converter(
            lambda x: xml.dom.minidom.parseString(x).toprettyxml(),
            _("format xml"),
        )


class _GeneratorLeaf(RunnableLeaf):
    def __init__(
        self,
        name: str,
        func: ty.Callable[[], str],
        description: str | None = None,
    ) -> None:
        super().__init__(func, name=name)
        self.description = description

    def has_result(self):
        return True

    def run(self, ctx=None):
        try:
            if res := self.object():
                return TextLeaf(res)

        except Exception as err:
            raise OperationError(f"Generator error: {err}") from err

        return None

    def get_description(self):
        return self.description

    def get_icon(self):
        return icons.ComposedIcon("stock_new-text", "kupfer-execute")


def _generate_alfanum_token(size: int) -> str:
    alphabet = string.ascii_letters + string.digits
    return "".join(secrets.choice(alphabet) for i in range(size))


class Generators(Source):
    serializable = None

    def __init__(self, name=_("Generators")):
        super().__init__(name)

    def is_dynamic(self):
        return True

    def get_items(self):
        for size in (16, 32, 64):
            yield _GeneratorLeaf(
                _("Random %(size)d-bytes hex token") % {"size": size},
                partial(secrets.token_hex, size),
            )

        for size in (16, 32, 64):
            yield _GeneratorLeaf(
                _("Random %(size)d-bytes alpha-numeric token")
                % {"size": size},
                partial(_generate_alfanum_token, size),
            )

        yield _GeneratorLeaf(
            _("UUID based on host and time"),
            lambda: str(uuid.uuid1()),
        )
        yield _GeneratorLeaf(_("Random UUID"), lambda: str(uuid.uuid4()))

        yield _GeneratorLeaf(
            _("Current time in ISO8601 format"),
            datetime.datetime.now().isoformat,
        )
        yield _GeneratorLeaf(
            _("Current time as Unix timestamp"),
            lambda: str(int(datetime.datetime.now().timestamp())),
        )
        yield _GeneratorLeaf(
            _("Current time as timestamp"),
            lambda: str(int(datetime.datetime.now().timestamp() * 1000)),
        )

    def provides(self):
        yield RunnableLeaf

    def get_gicon(self):
        return icons.ComposedIcon("stock_new-text", "kupfer-execute")

    def get_icon_name(self):
        return "stock_new-text"

    def get_description(self):
        return _("Generate text content with selected tool")



#! /usr/bin/env python3

"""
Show application help.
"""

from __future__ import annotations

import typing as ty

from kupfer import launch, version

if ty.TYPE_CHECKING:
    from kupfer.ui.uievents import GUIEnvironmentContext


def show_help(ctxenv: GUIEnvironmentContext | None = None) -> None:
    """Show Kupfer help pages, if possible"""
    if not launch.show_help_url(f"help:{version.PACKAGE_NAME}"):
        launch.show_url(version.HELP_WEBSITE)



"""
Global actions.

This file is a part of the program kupfer, which is
released under GNU General Public License v3 (or any later version),
see the main program file, and COPYING for details.
"""

from __future__ import annotations

import typing as ty

from kupfer import launch, support
from kupfer.desktop_launch import SpawnError
from kupfer.obj.base import Action, Leaf
from kupfer.obj.exceptions import OperationError
from kupfer.obj.objects import RunnableLeaf

if ty.TYPE_CHECKING:
    from gettext import gettext as _

    from kupfer.core import commandexec

__all__ = ("Execute", "OpenTerminal", "OpenUrl", "Perform")


class OpenTerminal(Action):
    action_accelerator = "t"

    def __init__(self, name: str = _("Open Terminal Here")) -> None:
        super().__init__(name)

    def wants_context(self) -> bool:
        return True

    def activate(
        self,
        leaf: Leaf,
        iobj: Leaf | None = None,
        ctx: commandexec.ExecutionToken | None = None,
    ) -> Leaf | None:
        assert ctx
        try:
            launch.spawn_terminal(leaf.object, ctx.environment.get_screen())
            return None
        except SpawnError as exc:
            raise OperationError(exc) from exc

    def get_description(self) -> str | None:
        return _("Open this location in a terminal")

    def get_icon_name(self) -> str:
        return "utilities-terminal"


class Execute(Action):
    """Execute executable file (FileLeaf)"""

    rank_adjust = 10

    def __init__(self, in_terminal=False, quoted=True):
        name = _("Run in Terminal") if in_terminal else _("Run (Execute)")
        super().__init__(name)
        self.in_terminal = in_terminal
        self.quoted = quoted

    def repr_key(self):
        return (self.in_terminal, self.quoted)

    def activate(
        self,
        leaf: Leaf,
        iobj: Leaf | None = None,
        ctx: commandexec.ExecutionToken | None = None,
    ) -> Leaf | None:
        if self.quoted:
            argv = [leaf.object]
        else:
            argv = support.argv_for_commandline(leaf.object)
        if self.in_terminal:
            launch.spawn_in_terminal(argv)
        else:
            launch.spawn_async(argv)

        return None

    def get_description(self) -> str | None:
        if self.in_terminal:
            return _("Run this program in a Terminal")

        return _("Run this program")


class OpenUrl(Action):
    action_accelerator: str | None = "o"
    rank_adjust: int = 5

    def __init__(self, name: str = _("Open URL")) -> None:
        super().__init__(name)

    def activate(
        self,
        leaf: Leaf,
        iobj: Leaf | None = None,
        ctx: commandexec.ExecutionToken | None = None,
    ) -> Leaf | None:
        url = leaf.object
        self.open_url(url)
        return None

    def open_url(self, url: str) -> None:
        launch.show_url(url)

    def get_description(self) -> str:
        return _("Open URL with default viewer")

    def get_icon_name(self) -> str:
        return "forward"


class Perform(Action):
    """Perform the action in a RunnableLeaf.

    RunnableLeaf can return result. In this case `has_result` and `item_types`
    must be specified.
    """

    action_accelerator: str | None = "o"
    rank_adjust = 5

    def __init__(
        self,
        name: str = _("Run"),
        has_result: bool = False,
        item_types: ty.Collection[Leaf] = (),
    ):
        # TRANS: 'Run' as in Perform a (saved) command
        super().__init__(name=name)
        self._has_result = has_result
        self._item_types = item_types

    def has_result(self):
        return self._has_result

    def wants_context(self) -> bool:
        return True

    def activate(
        self,
        leaf: Leaf,
        iobj: Leaf | None = None,
        ctx: commandexec.ExecutionToken | None = None,
    ) -> Leaf | None:
        assert isinstance(leaf, RunnableLeaf)
        if leaf.wants_context():
            return ty.cast("ty.Optional[Leaf]", leaf.run(ctx=ctx))

        return ty.cast("ty.Optional[Leaf]", leaf.run())

    def get_description(self) -> str:
        return _("Perform command")

    def item_types(self):
        return self._item_types



"""Persistent Globally Unique Indentifiers for KupferObjects.

Some objects are assigned identifiers by reference, some are assigned
identifiers containing the whole object data (SerializedObject).

SerializedObject is a saved representation of a KupferObject, i.e. a
data model user-level object.

We unpickle SerializedObjects in an especially conservative way: new
module loading is always refused; this way, we avoid loading parts of
the program that we didn't wish to activate.
"""

from __future__ import annotations

import contextlib
import pickle
import typing as ty

from kupfer.core import actioncompat, qfurl
from kupfer.core.sources import get_source_controller
from kupfer.obj import Action, AnySource, Leaf, Source
from kupfer.support import conspickle, pretty

__all__ = (
    "SerializedObject",
    "get_unique_id",
    "is_reference",
    "resolve_action_id",
    "resolve_unique_id",
)


_SERIALIZABLE_ATTRIBUTE = "serializable"


class SerializedObject:
    # treat the serializable attribute as a version number, defined on the class
    def __init__(self, obj: Leaf) -> None:
        self.version = getattr(obj, _SERIALIZABLE_ATTRIBUTE)
        self.data = pickle.dumps(obj, pickle.HIGHEST_PROTOCOL)

    def __hash__(self) -> int:
        return hash(self.data)

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, type(self))
            and self.data == other.data
            and self.version == other.version
        )

    def reconstruct(self) -> Leaf:
        obj = conspickle.ConservativeUnpickler.loads(self.data)
        if self.version != getattr(obj, _SERIALIZABLE_ATTRIBUTE):
            raise ValueError(f"Version mismatch for reconstructed {obj}")

        assert isinstance(obj, Leaf)
        return obj


PuID = ty.Union[str, SerializedObject]  # pylint: disable=invalid-name


def get_unique_id(obj: ty.Any) -> PuID | None:
    """Return unique id for object: if obj has attribute  qf_id  - return URI to
    locate object in Kupfer catalog (as str); if object is serializable
    return SerializedObject for `obj`; otherwise return string representation
    of object.
    Return None if obj is None or create SerializedObject failed.
    """
    if obj is None:
        return None

    if hasattr(obj, "qf_id"):
        return str(qfurl.Qfurl(obj))

    if getattr(obj, _SERIALIZABLE_ATTRIBUTE, None) is not None:
        try:
            return SerializedObject(obj)
        except pickle.PicklingError as exc:
            pretty.print_error(__name__, type(exc).__name__, exc)
            return None

    return repr(obj)


def is_reference(puid: ty.Any) -> bool:
    """Return True if @puid is a reference-type ID"""
    return not isinstance(puid, SerializedObject)


# A Context manager to block recursion when seeking inside a
# catalog; we have a stack (@_EXCLUDING) of the sources we
# are visiting, and nested context with the _exclusion
# context manager

_EXCLUDING: ty.Final[list[AnySource]] = []


@contextlib.contextmanager
def _exclusion(src):
    try:
        _EXCLUDING.append(src)
        yield
    finally:
        _EXCLUDING.pop()


def _is_currently_excluding(src: ty.Any) -> bool:
    return src is not None and src in _EXCLUDING


def _find_obj_in_catalog(
    puid: str, catalog: ty.Collection[Source]
) -> Leaf | None:
    if puid.startswith(qfurl.QFURL_SCHEME):
        qfu = qfurl.Qfurl(url=puid)
        return qfu.resolve_in_catalog(catalog)

    for src in catalog:
        if _is_currently_excluding(src):
            continue

        with _exclusion(src):
            for obj in src.get_leaves() or []:
                if repr(obj) == puid:
                    return obj

    return None


def resolve_unique_id(
    puid: ty.Any, excluding: AnySource | None = None
) -> Leaf | Action | None:
    """Resolve unique id @puid.

    The caller (if a Source) should pass itself as @excluding,
    so that recursion into itself is avoided.
    """
    if excluding is not None:
        with _exclusion(excluding):
            return resolve_unique_id(puid, None)

    if puid is None:
        return None

    if isinstance(puid, SerializedObject):
        try:
            return puid.reconstruct()
        except Exception as exc:
            pretty.print_debug(__name__, type(exc).__name__, exc)
            return None

    sctl = get_source_controller()
    if (obj := _find_obj_in_catalog(puid, sctl.firstlevel)) is not None:
        return obj

    other_sources = sctl.get_sources().copy() - set(sctl.firstlevel)
    return _find_obj_in_catalog(puid, other_sources)


def resolve_action_id(
    puid: ty.Any, for_item: Leaf | None = None
) -> Action | None:
    if puid is None:
        return None

    if isinstance(puid, SerializedObject):
        act = resolve_unique_id(puid)
        assert act is None or isinstance(act, Action)
        return act

    sctr = get_source_controller()
    if for_item is not None:
        for action in actioncompat.actions_for_item(for_item, sctr):
            if get_unique_id(action) == puid:
                return action

    get_action_id = repr
    for actions in sctr.action_decorators.values():
        for action in actions:
            if get_action_id(action) == puid:
                return action

    pretty.print_debug(__name__, f"Unable to resolve {puid} ({for_item})")
    return None



__kupfer_name__ = _("Devhelp")
__kupfer_actions__ = ("LookUp",)
__description__ = _("Search in Devhelp")
__version__ = "2017.1"
__author__ = ""
import typing as ty

from kupfer import icons, launch, plugin_support
from kupfer.obj import Action, OperationError, TextLeaf

if ty.TYPE_CHECKING:
    from gettext import gettext as _

plugin_support.check_command_available("devhelp")


class LookUp(Action):
    fallback_icon_name = "edit-find"

    def __init__(self):
        Action.__init__(self, _("Search in Devhelp"))

    def activate(self, leaf, iobj=None, ctx=None):
        text = leaf.object
        try:
            launch.spawn_async_raise(["devhelp", f"--search={text}"])
        except launch.SpawnError as exc:
            raise OperationError(exc) from exc

    def item_types(self):
        yield TextLeaf

    def valid_for_item(self, leaf):
        text = leaf.object
        return "\n" not in text

    def get_icon_name(self):
        return "devhelp"

    def get_gicon(self):
        return icons.ComposedIcon("devhelp", "edit-find")



from __future__ import annotations

import operator
import typing as ty

from kupfer.core import search
from kupfer.core.search import Rankable
from kupfer.support import pretty
from kupfer.support.itertools import peekfirst, unique_iterator

if ty.TYPE_CHECKING:
    from kupfer.obj.base import Action, Leaf, Source, TextSource

__all__ = ("Searcher",)

T = ty.TypeVar("T")
# function that validate leaves before search
ItemCheckFunc = ty.Callable[[ty.Iterable[T]], ty.Iterable[T]]
# function that decorate leaves before access
DecoratorFunc = ty.Callable[[ty.Iterable[Rankable]], ty.Iterable[Rankable]]


def _identity(x: ty.Any) -> ty.Any:
    return x


def _as_set_iter(seq: ty.Iterable[Rankable]) -> ty.Iterable[Rankable]:
    key = operator.attrgetter("object")
    return unique_iterator(seq, key=key)


def _valid_check(seq: ty.Iterable[Rankable]) -> ty.Iterable[Rankable]:
    """yield items of @seq that are valid"""
    for itm in seq:
        obj = itm.object
        if (not hasattr(obj, "is_valid")) or obj.is_valid():
            yield itm


class Searcher:
    """This class searches KupferObjects efficiently, and
    stores searches in a cache for a very limited time (*).

    (*) As of this writing, the cache is used when the old key
    is a prefix of the search key.
    """

    __slots__ = ("_old_key", "_source_cache")

    def __init__(self):
        self._source_cache = {}
        self._old_key: str | None = None

    def reset(self):
        self._source_cache.clear()
        self._old_key = None

    # pylint: disable=too-many-locals,too-many-branches
    def search(
        self,
        sources_: ty.Iterable[Source | TextSource],
        key: str,
        score: bool = True,
        item_check: ItemCheckFunc[Leaf | Action] | None = None,
        decorator: DecoratorFunc | None = None,
    ) -> tuple[Rankable | None, ty.Iterable[Rankable]]:
        """
        @sources is a sequence listing the inputs, which should be
        Sources, TextSources.

        If @score, sort by rank.
        filters (with _identity() as default):
            @item_check: Check items before adding to search pool
            @decorator: Decorate items before access

        Return (first, match_iter), where first is the first match,
        and match_iter an iterator to all matches, including the first match.
        """
        # use lowercase for search, but for text sources keep original case.
        keyl = key.lower()

        if not self._old_key or not keyl.startswith(self._old_key):
            self._source_cache.clear()

        self._old_key = keyl

        # General strategy: Extract a `list` from each source,
        # and perform ranking as in place operations on lists
        item_check = item_check or _identity
        decorator = decorator or _identity
        start_time = pretty.timing_start()
        match_lists: list[Rankable] = []
        for src in sources_:
            fixedrank = 0
            can_cache = True
            src_hash = hash(src)
            # Look in source cache for stored rankables
            rankables = self._source_cache.get(src_hash)
            if not rankables:
                if hasattr(src, "get_text_items"):
                    # TextSources
                    items = src.get_text_items(key)
                    fixedrank = src.get_rank()  # type: ignore
                    can_cache = False
                else:
                    # Source
                    items = src.get_leaves()

                rankables = search.make_rankables(item_check(items))

            if not rankables:
                continue

            if score:
                if fixedrank:
                    rankables = search.add_rank_objects(rankables, fixedrank)
                elif keyl:
                    rankables = search.bonus_objects(
                        search.score_objects(rankables, keyl),
                        keyl,
                        src.rank_adjust,
                    )

                if can_cache:
                    rankables = tuple(rankables)
                    self._source_cache[src_hash] = rankables

            match_lists.extend(rankables)

        matches = search.find_best_sort(match_lists) if score else match_lists

        # Check if the items are valid as the search
        # results are accessed through the iterators
        unique_matches = _as_set_iter(matches)
        match, match_iter = peekfirst(decorator(_valid_check(unique_matches)))
        pretty.timing_step(__name__, start_time, "ranked")
        return match, match_iter

    def rank_actions(
        self,
        objects: ty.Iterable[Action],
        key: str,
        leaf: Leaf | None,
        item_check: ItemCheckFunc[Action] | None = None,
        decorator: DecoratorFunc | None = None,
    ) -> tuple[Rankable | None, ty.Iterable[Rankable]]:
        """Rank actions by `key` for `leaf`.

        rank @objects, which should be a sequence of KupferObjects,
        for @key, with the action ranker algorithm.

        @leaf is the Leaf the action is going to be invoked on

        Filters and return value like .score().
        """
        item_check = item_check or _identity
        decorator = decorator or _identity

        rankables = search.make_rankables(item_check(objects))
        if key:
            key = key.lower()
            rankables = search.score_objects(rankables, key)
            matches = search.bonus_actions(rankables, key)
        else:
            matches = search.score_actions(rankables, leaf)

        sorted_matches = sorted(
            matches, key=operator.attrgetter("rank"), reverse=True
        )
        match, match_iter = peekfirst(decorator(sorted_matches))
        return match, match_iter



__kupfer_name__ = _("Locate Files")
__kupfer_actions__ = ("Locate",)
__description__ = _("Search filesystem using locate")
__version__ = ""
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import shutil
import subprocess
import typing as ty

from kupfer import icons, plugin_support
from kupfer.obj import Action, Leaf, Source, TextLeaf
from kupfer.obj.filesrc import construct_file_leaf
from kupfer.obj.special import CommandNotAvailableLeaf
from kupfer.support import kupferstring

if ty.TYPE_CHECKING:
    from gettext import gettext as _

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "ignore_case",
        "label": _("Ignore case distinctions when searching files"),
        "type": bool,
        "value": True,
    },
)

plugin_support.check_command_available("locate")


class Locate(Action):
    rank_adjust: int = -5

    def __init__(self):
        Action.__init__(self, _("Locate Files"))

    def is_factory(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        return LocateQuerySource(leaf.object)

    def item_types(self):
        yield TextLeaf

    def get_description(self):
        return _("Search filesystem using locate")

    def get_gicon(self):
        return icons.ComposedIcon("gnome-terminal", self.get_icon_name())

    def get_icon_name(self):
        return "edit-find"


class LocateQuerySource(Source):
    def __init__(self, query):
        Source.__init__(self, name=_('Results for "%s"') % query)
        self.query = query
        self.max_items = 500

    def repr_key(self):
        return self.query

    def get_items(self):
        locate_cmd = shutil.which("locate")
        if not locate_cmd:
            yield CommandNotAvailableLeaf(__name__, __kupfer_name__, "locate")
            return

        ignore_case = (
            "--ignore-case" if __kupfer_settings__["ignore_case"] else ""
        )
        # Start two processes, one to take the first hits, one
        # to take the remaining up to maximum. We start both at the same time
        # (regrettably, locate wont output streamingly to stdout)
        # but we ask the second for results only after iterating the first few
        first_num = 12

        def load(limit: int, offset: int) -> ty.Iterator[Leaf]:
            command = [
                locate_cmd,
                "--null",
                "--limit",
                str(limit),
                ignore_case,
                self.query,
            ]
            with subprocess.Popen(command, stdout=subprocess.PIPE) as proc:
                out, ignored_err = proc.communicate()
                for f in out.split(b"\x00")[offset:-1]:
                    yield construct_file_leaf(kupferstring.fromlocale(f))

        yield from load(first_num, 0)
        yield from load(self.max_items, first_num)

    def get_gicon(self):
        return icons.ComposedIcon("gnome-terminal", self.get_icon_name())

    def get_icon_name(self):
        return "edit-find"



"""
Support function for iterators

This file is a part of the program kupfer, which is
released under GNU General Public License v3 (or any later version),
see the main program file, and COPYING for details.
"""

from __future__ import annotations

import itertools
import typing as ty

__all__ = (
    "SavedIterable",
    "as_list",
    "peekfirst",
    "two_part_mapper",
    "unique_iterator",
)


def two_part_mapper(
    instr: str, repfunc: ty.Callable[[str], str | None]
) -> str:
    """Scan @instr two characters at a time and replace using @repfunc.
    If @repfunc return not None - use origin character.
    """
    if not instr:
        return instr

    def _inner():
        sit = zip(instr, instr[1:])
        for cur, nex in sit:
            key = cur + nex
            if (rep := repfunc(key)) is not None:
                yield rep
                # skip a step in the iter
                try:
                    next(sit)
                except StopIteration:
                    return

            else:
                yield cur

        yield instr[-1]

    return "".join(_inner())


T = ty.TypeVar("T")


def peekfirst(
    seq: ty.Iterable[T],
) -> tuple[T | None, ty.Iterable[T]]:
    """This function will return (firstitem, iter) where firstitem is the first
    item of `seq` or None if empty, and iter an equivalent copy of `seq`
    """
    seq = iter(seq)
    for itm in seq:
        old_iter = itertools.chain((itm,), seq)
        return (itm, old_iter)

    return (None, seq)


class SavedIterable(ty.Generic[T]):
    """Wrap an iterable and cache it.

    The SavedIterable can be accessed streamingly, while still being
    incrementally cached. Later attempts to iterate it will access the
    whole of the sequence.

    When it has been cached to its full extent once, it reduces to a
    thin wrapper of a sequence iterator. The SavedIterable will pickle
    into a list.

    >>> s = SavedIterable(range(5))
    >>> next(iter(s))
    0
    >>> list(s)
    [0, 1, 2, 3, 4]

    >>> iter(s)   # doctest: +ELLIPSIS
    <list_iterator object at 0x...>

    >>> import pickle
    >>> pickle.loads(pickle.dumps(s))
    [0, 1, 2, 3, 4]

    >>> u = SavedIterable(list(range(5)))
    >>> one, two = iter(u), iter(u)
    >>> next(one), next(two)
    (0, 0)
    >>> list(two)
    [1, 2, 3, 4]
    >>> list(one)
    [1, 2, 3, 4]

    >>> SavedIterable(list(range(3)))
    [0, 1, 2]
    """

    def __new__(cls, iterable: ty.Iterable[T]) -> ty.Any:
        if isinstance(iterable, list):
            return iterable

        return object.__new__(cls)

    def __init__(self, iterable: ty.Iterable[T]) -> None:
        # parent iterator
        self._iterator: ty.Iterator[T] | None = iter(iterable)
        # cached data
        self._data: list[T] = []

    def __iter__(self) -> ty.Iterator[T]:
        if self._iterator is None:
            # iterator was exhausted, so we return data from cache
            return iter(self._data)

        return self._incremental_caching_iter()

    def _incremental_caching_iter(self) -> ty.Iterator[T]:
        """Get next item from local cache, if no more items is in cache, try
        to load from origin iterator."""
        indices = itertools.count()
        while True:
            idx = next(indices)
            try:
                yield self._data[idx]
            except IndexError:
                pass
            else:
                continue

            if self._iterator is None:
                return

            try:
                x = next(self._iterator)
                self._data.append(x)
                yield x
            except StopIteration:
                self._iterator = None

    def __reduce__(self) -> tuple[ty.Any, ...]:
        # pickle into a list with __reduce__
        # (callable, args, state, listitems)
        return (list, (), None, iter(self))


K = ty.TypeVar("K")


def unique_iterator(
    seq: ty.Iterable[T],
    key: ty.Callable[[T], K] | None = None,
) -> ty.Iterator[T]:
    """Yield items of `seq` with set semantics; no duplicates.
    If `key` is given, value of key(object) is used for detecting duplicates.

    >>> list(unique_iterator([1, 2, 3, 3, 5, 1]))
    [1, 2, 3, 5]
    >>> list(unique_iterator([1, -2, 3, -3, -5, 2], key=abs))
    [1, -2, 3, -5]
    """
    if key is None:
        coll: set[T] = set()
        # itertools is ~10% faster
        for obj in itertools.filterfalse(coll.__contains__, seq):
            yield obj
            coll.add(obj)

    else:
        collk: set[K] = set()
        for obj in seq:
            if (value := key(obj)) not in collk:
                yield obj
                collk.add(value)


def as_list(seq: ty.Iterable[T]) -> ty.Collection[T]:
    """Return a list out of @seq, or seq if it is a list"""
    if isinstance(seq, (list, tuple)):
        return seq

    return list(seq)


if __name__ == "__main__":
    import doctest

    doctest.testmod()
